# -*- coding: utf-8 -*-
"""
Partial (Fused) Gromov-Wasserstein solvers.
"""

# Author: Laetitia Chapel <laetitia.chapel@irisa.fr>
#         Cédric Vincent-Cuaz <cedvincentcuaz@gmail.com>
#         Yikun Bai < yikun.bai@vanderbilt.edu >
#
# License: MIT License

from ..utils import list_to_array, unif
from ..backend import get_backend, NumpyBackend
from ..partial import entropic_partial_wasserstein
from ._utils import _transform_matrix, gwloss, gwggrad
from ..optim import partial_cg, solve_1d_linesearch_quad

import numpy as np
import warnings


def partial_gromov_wasserstein(
    C1,
    C2,
    p=None,
    q=None,
    m=None,
    loss_fun="square_loss",
    nb_dummies=1,
    G0=None,
    thres=1,
    numItermax=1e4,
    tol=1e-8,
    symmetric=None,
    warn=True,
    log=False,
    verbose=False,
    **kwargs,
):
    r"""
    Returns the Partial Gromov-Wasserstein transport between :math:`(\mathbf{C_1}, \mathbf{p})`
    and :math:`(\mathbf{C_2}, \mathbf{q})`.

    The function solves the following optimization problem using Conditional Gradient:

    .. math::
        \mathbf{T}^* \in \mathop{\arg \min}_\mathbf{T} \quad \sum_{i,j,k,l}
        L(\mathbf{C_1}_{i,k}, \mathbf{C_2}_{j,l}) \mathbf{T}_{i,j} \mathbf{T}_{k,l}

        s.t. \ \mathbf{T} \mathbf{1} &= \mathbf{p}

             \mathbf{T}^T \mathbf{1} &= \mathbf{q}

             \mathbf{T} &\geq 0

             \mathbf{1}^T \mathbf{T}^T \mathbf{1} = m &\leq \min\{\|\mathbf{p}\|_1, \|\mathbf{q}\|_1\}

    where :

    - :math:`\mathbf{C_1}`: Metric cost matrix in the source space.
    - :math:`\mathbf{C_2}`: Metric cost matrix in the target space.
    - :math:`\mathbf{p}`: Distribution in the source space.
    - :math:`\mathbf{q}`: Distribution in the target space.
    - `m` is the amount of mass to be transported
    - `L`: Loss function to account for the misfit between the similarity matrices.

    The formulation of the problem has been proposed in
    :ref:`[29] <references-partial-gromov-wasserstein>`

    .. note:: This function is backend-compatible and will work on arrays
        from all compatible backends. But the algorithm uses the C++ CPU backend
        which can lead to copy overhead on GPU arrays.
    .. note:: All computations in the conjugate gradient solver are done with
        numpy to limit memory overhead.
    .. note:: This function will cast the computed transport plan to the data
        type of the provided input :math:`\mathbf{C}_1`. Casting to an integer
        tensor might result in a loss of precision. If this behaviour is
        unwanted, please make sure to provide a floating point input.

    Parameters
    ----------
    C1 : array-like, shape (ns, ns)
        Metric cost matrix in the source space
    C2 : array-like, shape (nt, nt)
        Metric costfr matrix in the target space
    p : array-like, shape (ns,), optional
        Distribution in the source space.
        If let to its default value None, uniform distribution is taken.
    q : array-like, shape (nt,), optional
        Distribution in the target space.
        If let to its default value None, uniform distribution is taken.
    m : float, optional
        Amount of mass to be transported
        (default: :math:`\min\{\|\mathbf{p}\|_1, \|\mathbf{q}\|_1\}`)
    loss_fun : str, optional
        Loss function used for the solver either 'square_loss' or 'kl_loss'.
    nb_dummies : int, optional
        Number of dummy points to add (avoid instabilities in the EMD solver)
    G0 : array-like, shape (ns, nt), optional
        Initialization of the transportation matrix
    thres : float, optional
        quantile of the gradient matrix to populate the cost matrix when 0
        (default: 1)
    numItermax : int, optional
        Max number of iterations
    tol : float, optional
        tolerance for stopping iterations
    symmetric : bool, optional
        Either C1 and C2 are to be assumed symmetric or not.
        If let to its default None value, a symmetry test will be conducted.
        Else if set to True (resp. False), C1 and C2 will be assumed symmetric (resp. asymmetric).
    warn: bool, optional.
        Whether to raise a warning when EMD did not converge.
    log : bool, optional
        return log if True
    verbose : bool, optional
        Print information along iterations
    **kwargs : dict
        parameters can be directly passed to the emd solver


    Returns
    -------
    T : array-like, shape (`ns`, `nt`)
        Optimal transport matrix between the two spaces.

    log : dict
        Convergence information and loss.

    Examples
    --------
    >>> from ot.gromov import partial_gromov_wasserstein
    >>> import scipy as sp
    >>> a = np.array([0.25] * 4)
    >>> b = np.array([0.25] * 4)
    >>> x = np.array([1,2,100,200]).reshape((-1,1))
    >>> y = np.array([3,2,98,199]).reshape((-1,1))
    >>> C1 = sp.spatial.distance.cdist(x, x)
    >>> C2 = sp.spatial.distance.cdist(y, y)
    >>> np.round(partial_gromov_wasserstein(C1, C2, a, b),2)
    array([[0.  , 0.25, 0.  , 0.  ],
           [0.25, 0.  , 0.  , 0.  ],
           [0.  , 0.  , 0.25, 0.  ],
           [0.  , 0.  , 0.  , 0.25]])
    >>> np.round(partial_gromov_wasserstein(C1, C2, a, b, m=0.25),2)
    array([[0.  , 0.  , 0.  , 0.  ],
           [0.  , 0.  , 0.  , 0.  ],
           [0.  , 0.  , 0.25, 0.  ],
           [0.  , 0.  , 0.  , 0.  ]])


    .. _references-partial-gromov-wasserstein:
    References
    ----------
    ..  [29] Chapel, L., Alaya, M., Gasso, G. (2020). "Partial Optimal
        Transport with Applications on Positive-Unlabeled Learning".
        NeurIPS.

    """
    arr = [C1, C2]
    if p is not None:
        arr.append(list_to_array(p))
    else:
        p = unif(C1.shape[0], type_as=C1)
    if q is not None:
        arr.append(list_to_array(q))
    else:
        q = unif(C2.shape[0], type_as=C1)
    if G0 is not None:
        G0_ = G0
        arr.append(G0)

    nx = get_backend(*arr)
    p0, q0, C10, C20 = p, q, C1, C2

    p = nx.to_numpy(p0)
    q = nx.to_numpy(q0)
    C1 = nx.to_numpy(C10)
    C2 = nx.to_numpy(C20)
    if symmetric is None:
        symmetric = np.allclose(C1, C1.T, atol=1e-10) and np.allclose(
            C2, C2.T, atol=1e-10
        )

    if m is None:
        m = min(np.sum(p), np.sum(q))
    elif m < 0:
        raise ValueError("Problem infeasible. Parameter m should be greater" " than 0.")
    elif m > min(np.sum(p), np.sum(q)):
        raise ValueError(
            "Problem infeasible. Parameter m should lower or"
            " equal than min(|a|_1, |b|_1)."
        )

    if G0 is None:
        G0 = (
            np.outer(p, q) * m / (np.sum(p) * np.sum(q))
        )  # make sure |G0|=m, G01_m\leq p, G0.T1_n\leq q.

    else:
        G0 = nx.to_numpy(G0_)
        # Check marginals of G0
        assert np.all(G0.sum(1) <= p)
        assert np.all(G0.sum(0) <= q)

    q_extended = np.append(q, [(np.sum(p) - m) / nb_dummies] * nb_dummies)
    p_extended = np.append(p, [(np.sum(q) - m) / nb_dummies] * nb_dummies)

    # cg for GW is implemented using numpy on CPU
    np_ = NumpyBackend()

    fC1, fC2, hC1, hC2 = _transform_matrix(C1, C2, loss_fun, np_)
    fC2t = fC2.T
    if not symmetric:
        fC1t, hC1t, hC2t = fC1.T, hC1.T, hC2.T

    ones_p = np_.ones(p.shape[0], type_as=p)
    ones_q = np_.ones(q.shape[0], type_as=q)

    def f(G):
        pG = G.sum(1)
        qG = G.sum(0)
        constC1 = np.outer(np.dot(fC1, pG), ones_q)
        constC2 = np.outer(ones_p, np.dot(qG, fC2t))
        return gwloss(constC1 + constC2, hC1, hC2, G, np_)

    if symmetric:

        def df(G):
            pG = G.sum(1)
            qG = G.sum(0)
            constC1 = np.outer(np.dot(fC1, pG), ones_q)
            constC2 = np.outer(ones_p, np.dot(qG, fC2t))
            return gwggrad(constC1 + constC2, hC1, hC2, G, np_)
    else:

        def df(G):
            pG = G.sum(1)
            qG = G.sum(0)
            constC1 = np.outer(np.dot(fC1, pG), ones_q)
            constC2 = np.outer(ones_p, np.dot(qG, fC2t))
            constC1t = np.outer(np.dot(fC1t, pG), ones_q)
            constC2t = np.outer(ones_p, np.dot(qG, fC2))

            return 0.5 * (
                gwggrad(constC1 + constC2, hC1, hC2, G, np_)
                + gwggrad(constC1t + constC2t, hC1t, hC2t, G, np_)
            )

    def line_search(cost, G, deltaG, Mi, cost_G, df_G, **kwargs):
        df_Gc = df(deltaG + G)
        return solve_partial_gromov_linesearch(
            G, deltaG, cost_G, df_G, df_Gc, M=0.0, reg=1.0, nx=np_, **kwargs
        )

    if not nx.is_floating_point(C10):
        warnings.warn(
            "Input structure matrix consists of integers. The transport plan will be "
            "casted accordingly, possibly resulting in a loss of precision. "
            "If this behaviour is unwanted, please make sure your input "
            "structure matrix consists of floating point elements.",
            stacklevel=2,
        )

    if log:
        res, log = partial_cg(
            p,
            q,
            p_extended,
            q_extended,
            0.0,
            1.0,
            f,
            df,
            G0,
            line_search,
            log=True,
            numItermax=numItermax,
            stopThr=tol,
            stopThr2=0.0,
            warn=warn,
            **kwargs,
        )
        log["partial_gw_dist"] = nx.from_numpy(log["loss"][-1], type_as=C10)
        return nx.from_numpy(res, type_as=C10), log
    else:
        return nx.from_numpy(
            partial_cg(
                p,
                q,
                p_extended,
                q_extended,
                0.0,
                1.0,
                f,
                df,
                G0,
                line_search,
                log=False,
                numItermax=numItermax,
                stopThr=tol,
                stopThr2=0.0,
                **kwargs,
            ),
            type_as=C10,
        )


def partial_gromov_wasserstein2(
    C1,
    C2,
    p=None,
    q=None,
    m=None,
    loss_fun="square_loss",
    nb_dummies=1,
    G0=None,
    thres=1,
    numItermax=1e4,
    tol=1e-7,
    symmetric=None,
    warn=False,
    log=False,
    verbose=False,
    **kwargs,
):
    r"""
    Returns the Partial Gromov-Wasserstein discrepancy between
    :math:`(\mathbf{C_1}, \mathbf{p})` and :math:`(\mathbf{C_2}, \mathbf{q})`.

    The function solves the following optimization problem using Conditional Gradient:

    .. math::
        \mathbf{PGW} = \mathop{\min}_\mathbf{T} \quad \sum_{i,j,k,l}
        L(\mathbf{C_1}_{i,k}, \mathbf{C_2}_{j,l}) \mathbf{T}_{i,j} \mathbf{T}_{k,l}

        s.t. \ \mathbf{T} \mathbf{1} &= \mathbf{p}

             \mathbf{T}^T \mathbf{1} &= \mathbf{q}

             \mathbf{T} &\geq 0

             \mathbf{1}^T \mathbf{T}^T \mathbf{1} = m &\leq \min\{\|\mathbf{p}\|_1, \|\mathbf{q}\|_1\}

    where :

    - :math:`\mathbf{C_1}`: Metric cost matrix in the source space.
    - :math:`\mathbf{C_2}`: Metric cost matrix in the target space.
    - :math:`\mathbf{p}`: Distribution in the source space.
    - :math:`\mathbf{q}`: Distribution in the target space.
    - `m` is the amount of mass to be transported
    - `L`: Loss function to account for the misfit between the similarity matrices.


    The formulation of the problem has been proposed in
    :ref:`[29] <references-partial-gromov-wasserstein2>`

    Note that when using backends, this loss function is differentiable wrt the
    matrices (C1, C2).

    .. note:: This function is backend-compatible and will work on arrays
        from all compatible backends. But the algorithm uses the C++ CPU backend
        which can lead to copy overhead on GPU arrays.
    .. note:: All computations in the conjugate gradient solver are done with
        numpy to limit memory overhead.
    .. note:: This function will cast the computed transport plan to the data
        type of the provided input :math:`\mathbf{C}_1`. Casting to an integer
        tensor might result in a loss of precision. If this behaviour is
        unwanted, please make sure to provide a floating point input.

    Parameters
    ----------
    C1 : ndarray, shape (ns, ns)
        Metric cost matrix in the source space
    C2 : ndarray, shape (nt, nt)
        Metric cost matrix in the target space
    p : ndarray, shape (ns,)
        Distribution in the source space
    q : ndarray, shape (nt,)
        Distribution in the target space
    m : float, optional
        Amount of mass to be transported
        (default: :math:`\min\{\|\mathbf{p}\|_1, \|\mathbf{q}\|_1\}`)
    loss_fun : str, optional
        Loss function used for the solver either 'square_loss' or 'kl_loss'.
    nb_dummies : int, optional
        Number of dummy points to add (avoid instabilities in the EMD solver)
    G0 : ndarray, shape (ns, nt), optional
        Initialization of the transportation matrix
    thres : float, optional
        quantile of the gradient matrix to populate the cost matrix when 0
        (default: 1)
    numItermax : int, optional
        Max number of iterations
    tol : float, optional
        tolerance for stopping iterations
    symmetric : bool, optional
        Either C1 and C2 are to be assumed symmetric or not.
        If let to its default None value, a symmetry test will be conducted.
        Else if set to True (resp. False), C1 and C2 will be assumed symmetric (resp. asymmetric).
    warn: bool, optional.
        Whether to raise a warning when EMD did not converge.
    log : bool, optional
        return log if True
    verbose : bool, optional
        Print information along iterations
    **kwargs : dict
        parameters can be directly passed to the emd solver


    .. warning::
        When dealing with a large number of points, the EMD solver may face
        some instabilities, especially when the mass associated to the dummy
        point is large. To avoid them, increase the number of dummy points
        (allows a smoother repartition of the mass over the points).


    Returns
    -------
    partial_gw_dist : float
        partial GW discrepancy
    log : dict
        log dictionary returned only if `log` is `True`


    Examples
    --------
    >>> from ot.gromov import partial_gromov_wasserstein2
    >>> import scipy as sp
    >>> a = np.array([0.25] * 4)
    >>> b = np.array([0.25] * 4)
    >>> x = np.array([1,2,100,200]).reshape((-1,1))
    >>> y = np.array([3,2,98,199]).reshape((-1,1))
    >>> C1 = sp.spatial.distance.cdist(x, x)
    >>> C2 = sp.spatial.distance.cdist(y, y)
    >>> np.round(partial_gromov_wasserstein2(C1, C2, a, b),2)
    3.38
    >>> np.round(partial_gromov_wasserstein2(C1, C2, a, b, m=0.25),2)
    0.0


    .. _references-partial-gromov-wasserstein2:
    References
    ----------
    ..  [29] Chapel, L., Alaya, M., Gasso, G. (2020). "Partial Optimal
        Transport with Applications on Positive-Unlabeled Learning".
        NeurIPS.

    """
    # simple get_backend as the full one will be handled in gromov_wasserstein
    nx = get_backend(C1, C2)

    # init marginals if set as None
    if p is None:
        p = unif(C1.shape[0], type_as=C1)
    if q is None:
        q = unif(C2.shape[0], type_as=C1)

    T, log_pgw = partial_gromov_wasserstein(
        C1,
        C2,
        p,
        q,
        m,
        loss_fun,
        nb_dummies,
        G0,
        thres,
        numItermax,
        tol,
        symmetric,
        warn,
        True,
        verbose,
        **kwargs,
    )

    log_pgw["T"] = T
    pgw = log_pgw["partial_gw_dist"]

    if loss_fun == "square_loss":
        gC1 = 2 * C1 * nx.outer(p, p) - 2 * nx.dot(T, nx.dot(C2, T.T))
        gC2 = 2 * C2 * nx.outer(q, q) - 2 * nx.dot(T.T, nx.dot(C1, T))
    elif loss_fun == "kl_loss":
        gC1 = nx.log(C1 + 1e-15) * nx.outer(p, p) - nx.dot(
            T, nx.dot(nx.log(C2 + 1e-15), T.T)
        )
        gC2 = -nx.dot(T.T, nx.dot(C1, T)) / (C2 + 1e-15) + nx.outer(q, q)

    pgw = nx.set_gradients(pgw, (C1, C2), (gC1, gC2))

    if log:
        return pgw, log_pgw
    else:
        return pgw


def solve_partial_gromov_linesearch(
    G,
    deltaG,
    cost_G,
    df_G,
    df_Gc,
    M,
    reg,
    alpha_min=None,
    alpha_max=None,
    nx=None,
    **kwargs,
):
    """
    Solve the linesearch in the FW iterations of partial (F)GW following eq.5 of :ref:`[29]`.

    Parameters
    ----------

    G : array-like, shape(ns,nt)
        The transport map at a given iteration of the FW
    deltaG : array-like (ns,nt)
        Difference between the optimal map `Gc` found by linearization in the
        FW algorithm and the value at a given iteration
    cost_G : float
        Value of the cost at `G`
    df_G : array-like (ns,nt)
        Gradient of the GW cost at `G`
    df_Gc : array-like (ns,nt)
        Gradient of the GW cost at `Gc`
    M : array-like (ns,nt)
        Cost matrix between the features.
    reg : float
        Regularization parameter.
    alpha_min : float, optional
        Minimum value for alpha
    alpha_max : float, optional
        Maximum value for alpha
    nx : backend, optional
        If let to its default value None, a backend test will be conducted.

    Returns
    -------
    alpha : float
        The optimal step size of the FW
    fc : int
        nb of function call. Useless here
    cost_G : float
        The value of the cost for the next iteration
    df_G : array-like (ns,nt)
        Updated gradient of the GW cost

    References
    ----------
    ..  [29] Chapel, L., Alaya, M., Gasso, G. (2020). "Partial Optimal
        Transport with Applications on Positive-Unlabeled Learning".
        NeurIPS.

    """
    if nx is None:
        if isinstance(M, int) or isinstance(M, float):
            nx = get_backend(G, deltaG, df_G, df_Gc)
        else:
            nx = get_backend(G, deltaG, df_G, df_Gc, M)

    df_deltaG = df_Gc - df_G
    cost_deltaG = 0.5 * nx.sum(df_deltaG * deltaG)

    a = reg * cost_deltaG
    # formula to check for partial FGW
    b = nx.sum(M * deltaG) + reg * nx.sum(df_G * deltaG)

    alpha = solve_1d_linesearch_quad(a, b)
    if alpha_min is not None or alpha_max is not None:
        alpha = np.clip(alpha, alpha_min, alpha_max)

    # the new cost is deduced from the line search quadratic function
    cost_G = cost_G + a * (alpha**2) + b * alpha

    # update the gradient for next cg iteration
    df_G = df_G + alpha * df_deltaG
    return alpha, 1, cost_G, df_G


def entropic_partial_gromov_wasserstein(
    C1,
    C2,
    p=None,
    q=None,
    reg=1.0,
    m=None,
    loss_fun="square_loss",
    G0=None,
    numItermax=1000,
    tol=1e-7,
    symmetric=None,
    log=False,
    verbose=False,
):
    r"""
    Returns the partial Gromov-Wasserstein transport between
    :math:`(\mathbf{C_1}, \mathbf{p})` and :math:`(\mathbf{C_2}, \mathbf{q})`

    The function solves the following optimization problem:

    .. math::
        \gamma = \mathop{\arg \min}_{\gamma} \quad \sum_{i,j,k,l}
        L(\mathbf{C_1}_{i,k}, \mathbf{C_2}_{j,l})\cdot
        \gamma_{i,j}\cdot\gamma_{k,l} + \mathrm{reg} \cdot\Omega(\gamma)

    .. math::
        s.t. \ \gamma &\geq 0

             \gamma \mathbf{1} &\leq \mathbf{a}

             \gamma^T \mathbf{1} &\leq \mathbf{b}

             \mathbf{1}^T \gamma^T \mathbf{1} = m
             &\leq \min\{\|\mathbf{a}\|_1, \|\mathbf{b}\|_1\}

    where :

    - :math:`\mathbf{C_1}` is the metric cost matrix in the source space
    - :math:`\mathbf{C_2}` is the metric cost matrix in the target space
    - :math:`\mathbf{p}` and :math:`\mathbf{q}` are the sample weights
    - `L`: quadratic loss function
    - :math:`\Omega` is the entropic regularization term,
      :math:`\Omega=\sum_{i,j} \gamma_{i,j}\log(\gamma_{i,j})`
    - `m` is the amount of mass to be transported

    The formulation of the GW problem has been proposed in
    :ref:`[12] <references-entropic-partial-gromov-wasserstein>` and the
    partial GW in :ref:`[29] <references-entropic-partial-gromov-wasserstein>`

    Parameters
    ----------
    C1 : array-like, shape (ns, ns)
        Metric cost matrix in the source space
    C2 : array-like, shape (nt, nt)
        Metric cost matrix in the target space
    p : array-like, shape (ns,), optional
        Distribution in the source space.
        If let to its default value None, uniform distribution is taken.
    q : array-like, shape (nt,), optional
        Distribution in the target space.
        If let to its default value None, uniform distribution is taken.
    reg: float, optional. Default is 1.
        entropic regularization parameter
    m : float, optional
        Amount of mass to be transported (default:
        :math:`\min\{\|\mathbf{p}\|_1, \|\mathbf{q}\|_1\}`)
    loss_fun : str, optional
        Loss function used for the solver either 'square_loss' or 'kl_loss'.
    G0 : array-like, shape (ns, nt), optional
        Initialization of the transportation matrix
    numItermax : int, optional
        Max number of iterations
    tol : float, optional
        Stop threshold on error (>0)
    symmetric : bool, optional
        Either C1 and C2 are to be assumed symmetric or not.
        If let to its default None value, a symmetry test will be conducted.
        Else if set to True (resp. False), C1 and C2 will be assumed symmetric (resp. asymmetric).
    log : bool, optional
        return log if True
    verbose : bool, optional
        Print information along iterations

    Examples
    --------
    >>> from ot.gromov import entropic_partial_gromov_wasserstein
    >>> import scipy as sp
    >>> a = np.array([0.25] * 4)
    >>> b = np.array([0.25] * 4)
    >>> x = np.array([1,2,100,200]).reshape((-1,1))
    >>> y = np.array([3,2,98,199]).reshape((-1,1))
    >>> C1 = sp.spatial.distance.cdist(x, x)
    >>> C2 = sp.spatial.distance.cdist(y, y)
    >>> np.round(entropic_partial_gromov_wasserstein(C1, C2, a, b, 1e2), 2)
    array([[0.12, 0.13, 0.  , 0.  ],
           [0.13, 0.12, 0.  , 0.  ],
           [0.  , 0.  , 0.25, 0.  ],
           [0.  , 0.  , 0.  , 0.25]])
    >>> np.round(entropic_partial_gromov_wasserstein(C1, C2, a, b, 1e2,0.25), 2)
    array([[0.02, 0.03, 0.  , 0.03],
           [0.03, 0.03, 0.  , 0.03],
           [0.  , 0.  , 0.03, 0.  ],
           [0.02, 0.02, 0.  , 0.03]])

    Returns
    -------
    :math: `gamma` : (dim_a, dim_b) ndarray
        Optimal transportation matrix for the given parameters
    log : dict
        log dictionary returned only if `log` is `True`


    .. _references-entropic-partial-gromov-wasserstein:
    References
    ----------
    .. [12] Peyré, Gabriel, Marco Cuturi, and Justin Solomon,
        "Gromov-Wasserstein averaging of kernel and distance matrices."
        International Conference on Machine Learning (ICML). 2016.

    .. [29] Chapel, L., Alaya, M., Gasso, G. (2020). "Partial Optimal
        Transport with Applications on Positive-Unlabeled Learning".
        NeurIPS.

    See Also
    --------
    ot.partial.partial_gromov_wasserstein: exact Partial Gromov-Wasserstein
    """

    arr = [C1, C2, G0]
    if p is not None:
        p = list_to_array(p)
        arr.append(p)
    if q is not None:
        q = list_to_array(q)
        arr.append(q)

    nx = get_backend(*arr)

    if p is None:
        p = nx.ones(C1.shape[0], type_as=C1) / C1.shape[0]
    if q is None:
        q = nx.ones(C2.shape[0], type_as=C2) / C2.shape[0]

    if m is None:
        m = min(nx.sum(p), nx.sum(q))
    elif m < 0:
        raise ValueError("Problem infeasible. Parameter m should be greater" " than 0.")
    elif m > min(nx.sum(p), nx.sum(q)):
        raise ValueError(
            "Problem infeasible. Parameter m should lower or"
            " equal than min(|a|_1, |b|_1)."
        )

    if G0 is None:
        G0 = (
            nx.outer(p, q) * m / (nx.sum(p) * nx.sum(q))
        )  # make sure |G0|=m, G01_m\leq p, G0.T1_n\leq q.

    else:
        # Check marginals of G0
        assert nx.any(nx.sum(G0, 1) <= p)
        assert nx.any(nx.sum(G0, 0) <= q)

    if symmetric is None:
        symmetric = np.allclose(C1, C1.T, atol=1e-10) and np.allclose(
            C2, C2.T, atol=1e-10
        )

    # Setup gradient computation
    fC1, fC2, hC1, hC2 = _transform_matrix(C1, C2, loss_fun, nx)
    fC2t = fC2.T
    if not symmetric:
        fC1t, hC1t, hC2t = fC1.T, hC1.T, hC2.T

    ones_p = nx.ones(p.shape[0], type_as=p)
    ones_q = nx.ones(q.shape[0], type_as=q)

    def f(G):
        pG = nx.sum(G, 1)
        qG = nx.sum(G, 0)
        constC1 = nx.outer(nx.dot(fC1, pG), ones_q)
        constC2 = nx.outer(ones_p, nx.dot(qG, fC2t))
        return gwloss(constC1 + constC2, hC1, hC2, G, nx)

    if symmetric:

        def df(G):
            pG = nx.sum(G, 1)
            qG = nx.sum(G, 0)
            constC1 = nx.outer(nx.dot(fC1, pG), ones_q)
            constC2 = nx.outer(ones_p, nx.dot(qG, fC2t))
            return gwggrad(constC1 + constC2, hC1, hC2, G, nx)
    else:

        def df(G):
            pG = nx.sum(G, 1)
            qG = nx.sum(G, 0)
            constC1 = nx.outer(nx.dot(fC1, pG), ones_q)
            constC2 = nx.outer(ones_p, nx.dot(qG, fC2t))
            constC1t = nx.outer(nx.dot(fC1t, pG), ones_q)
            constC2t = nx.outer(ones_p, nx.dot(qG, fC2))

            return 0.5 * (
                gwggrad(constC1 + constC2, hC1, hC2, G, nx)
                + gwggrad(constC1t + constC2t, hC1t, hC2t, G, nx)
            )

    cpt = 0
    err = 1

    loge = {"err": []}

    while err > tol and cpt < numItermax:
        Gprev = G0
        M_entr = df(G0)
        G0 = entropic_partial_wasserstein(p, q, M_entr, reg, m)
        if cpt % 10 == 0:  # to speed up the computations
            err = np.linalg.norm(G0 - Gprev)
            if log:
                loge["err"].append(err)
            if verbose:
                if cpt % 200 == 0:
                    print(
                        "{:5s}|{:12s}|{:12s}".format("It.", "Err", "Loss")
                        + "\n"
                        + "-" * 31
                    )
                print("{:5d}|{:8e}|{:8e}".format(cpt, err, f(G0)))

        cpt += 1

    if log:
        loge["partial_gw_dist"] = f(G0)
        return G0, loge
    else:
        return G0


def entropic_partial_gromov_wasserstein2(
    C1,
    C2,
    p=None,
    q=None,
    reg=1.0,
    m=None,
    loss_fun="square_loss",
    G0=None,
    numItermax=1000,
    tol=1e-7,
    symmetric=None,
    log=False,
    verbose=False,
):
    r"""
    Returns the partial Gromov-Wasserstein discrepancy between
    :math:`(\mathbf{C_1}, \mathbf{p})` and :math:`(\mathbf{C_2}, \mathbf{q})`

    The function solves the following optimization problem:

    .. math::
        PGW = \min_{\gamma} \quad \sum_{i,j,k,l} L(\mathbf{C_1}_{i,k},
             \mathbf{C_2}_{j,l})\cdot
             \gamma_{i,j}\cdot\gamma_{k,l} + \mathrm{reg} \cdot\Omega(\gamma)

    .. math::
        s.t. \ \gamma &\geq 0

             \gamma \mathbf{1} &\leq \mathbf{a}

             \gamma^T \mathbf{1} &\leq \mathbf{b}

             \mathbf{1}^T \gamma^T \mathbf{1} = m &\leq \min\{\|\mathbf{a}\|_1, \|\mathbf{b}\|_1\}

    where :

    - :math:`\mathbf{C_1}` is the metric cost matrix in the source space
    - :math:`\mathbf{C_2}` is the metric cost matrix in the target space
    - :math:`\mathbf{p}` and :math:`\mathbf{q}` are the sample weights
    - `L`: Loss function to account for the misfit between the similarity matrices.
    - :math:`\Omega` is the entropic regularization term,
      :math:`\Omega=\sum_{i,j} \gamma_{i,j}\log(\gamma_{i,j})`
    - `m` is the amount of mass to be transported

    The formulation of the GW problem has been proposed in
    :ref:`[12] <references-entropic-partial-gromov-wasserstein2>` and the
    partial GW in :ref:`[29] <references-entropic-partial-gromov-wasserstein2>`


    Parameters
    ----------
    C1 : ndarray, shape (ns, ns)
        Metric cost matrix in the source space
    C2 : ndarray, shape (nt, nt)
        Metric cost matrix in the target space
    p : array-like, shape (ns,), optional
        Distribution in the source space.
        If let to its default value None, uniform distribution is taken.
    q : array-like, shape (nt,), optional
        Distribution in the target space.
        If let to its default value None, uniform distribution is taken.
    reg: float
        entropic regularization parameter
    m : float, optional
        Amount of mass to be transported (default:
        :math:`\min\{\|\mathbf{p}\|_1, \|\mathbf{q}\|_1\}`)
    loss_fun : str, optional
        Loss function used for the solver either 'square_loss' or 'kl_loss'.
    G0 : ndarray, shape (ns, nt), optional
        Initialization of the transportation matrix
    numItermax : int, optional
        Max number of iterations
    tol : float, optional
        Stop threshold on error (>0)
    symmetric : bool, optional
        Either C1 and C2 are to be assumed symmetric or not.
        If let to its default None value, a symmetry test will be conducted.
        Else if set to True (resp. False), C1 and C2 will be assumed symmetric (resp. asymmetric).
    log : bool, optional
        return log if True
    verbose : bool, optional
        Print information along iterations


    Returns
    -------
    partial_gw_dist: float
        Partial Gromov-Wasserstein distance
    log : dict
        log dictionary returned only if `log` is `True`

    Examples
    --------
    >>> from ot.gromov import entropic_partial_gromov_wasserstein2
    >>> import scipy as sp
    >>> a = np.array([0.25] * 4)
    >>> b = np.array([0.25] * 4)
    >>> x = np.array([1,2,100,200]).reshape((-1,1))
    >>> y = np.array([3,2,98,199]).reshape((-1,1))
    >>> C1 = sp.spatial.distance.cdist(x, x)
    >>> C2 = sp.spatial.distance.cdist(y, y)
    >>> np.round(entropic_partial_gromov_wasserstein2(C1, C2, a, b, 1e2), 2)
    3.75


    .. _references-entropic-partial-gromov-wasserstein2:
    References
    ----------
    .. [12] Peyré, Gabriel, Marco Cuturi, and Justin Solomon,
        "Gromov-Wasserstein averaging of kernel and distance matrices."
        International Conference on Machine Learning (ICML). 2016.

    ..  [29] Chapel, L., Alaya, M., Gasso, G. (2020). "Partial Optimal
        Transport with Applications on Positive-Unlabeled Learning".
        NeurIPS.
    """

    partial_gw, log_gw = entropic_partial_gromov_wasserstein(
        C1, C2, p, q, reg, m, loss_fun, G0, numItermax, tol, symmetric, True, verbose
    )

    log_gw["T"] = partial_gw

    if log:
        return log_gw["partial_gw_dist"], log_gw
    else:
        return log_gw["partial_gw_dist"]



# -*- coding: utf-8 -*-
"""
Unbalanced Co-Optimal Transport and Fused Unbalanced Gromov-Wasserstein solvers
"""

# Author: Quang Huy Tran <quang-huy.tran@univ-ubs.fr>
#         Alexis Thual <alexis.thual@cea.fr>
#
# License: MIT License

import warnings
from functools import partial
import ot
from ot.backend import get_backend
from ot.utils import list_to_array, get_parameter_pair
from ._utils import (
    fused_unbalanced_across_spaces_cost,
    uot_cost_matrix,
    uot_parameters_and_measures,
)


def fused_unbalanced_across_spaces_divergence(
    X,
    Y,
    wx_samp=None,
    wx_feat=None,
    wy_samp=None,
    wy_feat=None,
    reg_marginals=10,
    epsilon=0,
    reg_type="joint",
    divergence="kl",
    unbalanced_solver="sinkhorn",
    alpha=0,
    M_samp=None,
    M_feat=None,
    rescale_plan=True,
    init_pi=None,
    init_duals=None,
    max_iter=100,
    tol=1e-7,
    max_iter_ot=500,
    tol_ot=1e-7,
    log=False,
    verbose=False,
    **kwargs_solver,
):
    r"""Compute the fused unbalanced cross-spaces divergence between two matrices equipped
    with the distributions on rows and columns. We consider two cases of matrix:

    - (Squared) similarity matrix in Gromov-Wasserstein setting,
    whose rows and columns represent the samples.

    - Arbitrary-size matrix in Co-Optimal Transport setting,
    whose rows represent samples, and columns represent corresponding features/dimensions.

    More precisely, this function returns the sample and feature transport plans between
    :math:`(\mathbf{X}, \mathbf{w}_{xs}, \mathbf{w}_{xf})` and
    :math:`(\mathbf{Y}, \mathbf{w}_{ys}, \mathbf{w}_{yf})`,
    by solving the following problem using Block Coordinate Descent algorithm:

    .. math::

        \mathop{\arg \min}_{\mathbf{P}, \mathbf{Q}}
        &\quad \sum_{i,j,k,l}
        (\mathbf{X}_{i,k} - \mathbf{Y}_{j,l})^2 \mathbf{P}_{i,j} \mathbf{Q}_{k,l} \\
        &+ \rho_s \mathbf{Div}(\mathbf{P}_{\# 1} \mathbf{Q}_{\# 1}^T | \mathbf{w}_{xs} \mathbf{w}_{ys}^T)
        + \rho_f \mathbf{Div}(\mathbf{P}_{\# 2} \mathbf{Q}_{\# 2}^T | \mathbf{w}_{xf} \mathbf{w}_{yf}^T) \\
        &+ \alpha_s \sum_{i,j} \mathbf{P}_{i,j} \mathbf{M^{(s)}}_{i, j}
        + \alpha_f \sum_{k, l} \mathbf{Q}_{k,l} \mathbf{M^{(f)}}_{k, l}
        + \mathbf{Reg}(\mathbf{P}, \mathbf{Q})

    Where:

    - :math:`\mathbf{X}`: Source input (arbitrary-size) matrix
    - :math:`\mathbf{Y}`: Target input (arbitrary-size) matrix
    - :math:`\mathbf{M^{(s)}}`: Additional sample matrix
    - :math:`\mathbf{M^{(f)}}`: Additional feature matrix
    - :math:`\mathbf{w}_{xs}`: Distribution of the samples in the source space
    - :math:`\mathbf{w}_{xf}`: Distribution of the features in the source space
    - :math:`\mathbf{w}_{ys}`: Distribution of the samples in the target space
    - :math:`\mathbf{w}_{yf}`: Distribution of the features in the target space
    - :math:`\mathbf{Div}`: Either Kullback-Leibler divergence or half-squared L2 norm.
    - :math:`\mathbf{Reg}`: Regularizer for sample and feature couplings.

    We consider two types of regularizer:
        + Independent regularization used in unbalanced Co-Optimal Transport

        .. math::
            \mathbf{Reg}(\mathbf{P}, \mathbf{Q}) =
            \varepsilon_s \mathbf{Div}(\mathbf{P} | \mathbf{w}_{xs} \mathbf{w}_{ys}^T)
            + \varepsilon_f \mathbf{Div}(\mathbf{Q} | \mathbf{w}_{xf} \mathbf{w}_{yf}^T)

        + Joint regularization used in fused unbalanced Gromov-Wasserstein

        .. math::
            \mathbf{Reg}(\mathbf{P}, \mathbf{Q}) =
            \varepsilon \mathbf{Div}(\mathbf{P} \otimes \mathbf{Q} | (\mathbf{w}_{xs} \mathbf{w}_{ys}^T) \otimes (\mathbf{w}_{xf} \mathbf{w}_{yf}^T) )

    .. note:: This function allows epsilon to be zero. In that case, `unbalanced_method` must be either "mm" or "lbfgsb".

    Parameters
    ----------
    X : (n_sample_x, n_feature_x) array-like, float
        Source input matrix.
    Y : (n_sample_y, n_feature_y) array-like, float
        Target input matrix.
    wx_samp : (n_sample_x, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix X.
        Uniform distribution by default.
    wx_feat : (n_feature_x, ) array-like, float, optional (default = None)
        Histogram assigned on columns (features) of matrix X.
        Uniform distribution by default.
    wy_samp : (n_sample_y, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix Y.
        Uniform distribution by default.
    wy_feat : (n_feature_y, ) array-like, float, optional (default = None)
        Histogram assigned on columns (features) of matrix Y.
        Uniform distribution by default.
    reg_marginals: float or indexable object of length 1 or 2
        Marginal relaxation terms for sample and feature couplings.
        If `reg_marginals` is a scalar or an indexable object of length 1,
        then the same value is applied to both marginal relaxations.
    epsilon : scalar or indexable object of length 2, float or int, optional (default = 0)
        Regularization parameters for entropic approximation of sample and feature couplings.
        Allow the case where `epsilon` contains 0. In that case, the MM solver is used by default
        instead of Sinkhorn solver. If `epsilon` is scalar, then the same value is applied to
        both regularization of sample and feature couplings.
    reg_type: string, optional

        - If `reg_type` = "joint": then use joint regularization for couplings.

        - If `reg_type` = "independent": then use independent regularization for couplings.
    divergence : string, optional (default = "kl")

        - If `divergence` = "kl", then Div is the Kullback-Leibler divergence.

        - If `divergence` = "l2", then Div is the half squared Euclidean norm.
    unbalanced_solver : string, optional (default = "sinkhorn")
        Solver for the unbalanced OT subroutine.

        - If `divergence` = "kl", then `unbalanced_solver` can be: "sinkhorn", "sinkhorn_log", "mm", "lbfgsb"

        - If `divergence` = "l2", then `unbalanced_solver` can be "mm", "lbfgsb"
    alpha : scalar or indexable object of length 2, float or int, optional (default = 0)
        Coeffficient parameter of linear terms with respect to the sample and feature couplings.
        If alpha is scalar, then the same alpha is applied to both linear terms.
    M_samp : (n_sample_x, n_sample_y), float, optional (default = None)
        Sample matrix associated to the Wasserstein linear term on sample coupling.
    M_feat : (n_feature_x, n_feature_y), float, optional (default = None)
        Feature matrix associated to the Wasserstein linear term on feature coupling.
    rescale_plan : boolean, optional (default = True)
        If True, then rescale the sample and feature transport plans within each BCD iteration,
        so that they always have equal mass.
    init_pi : tuple of two matrices of size (n_sample_x, n_sample_y) and
        (n_feature_x, n_feature_y), optional (default = None).
        Initialization of sample and feature couplings.
        Uniform distributions by default.
    init_duals : tuple of two tuples ((n_sample_x, ), (n_sample_y, )) and ((n_feature_x, ), (n_feature_y, )), optional (default = None).
        Initialization of sample and feature dual vectors
        if using Sinkhorn algorithm. Zero vectors by default.
    max_iter : int, optional (default = 100)
        Number of Block Coordinate Descent (BCD) iterations.
    tol : float, optional (default = 1e-7)
        Tolerance of BCD scheme. If the L1-norm between the current and previous
        sample couplings is under this threshold, then stop BCD scheme.
    max_iter_ot : int, optional (default = 100)
        Number of iterations to solve each of the
        two unbalanced optimal transport problems in each BCD iteration.
    tol_ot : float, optional (default = 1e-7)
        Tolerance of unbalanced solver for each of the
        two unbalanced optimal transport problems in each BCD iteration.
    log : bool, optional (default = False)
        If True then the cost and four dual vectors, including
        two from sample and two from feature couplings, are recorded.
    verbose : bool, optional (default = False)
        If True then print the COOT cost at every multiplier of `eval_bcd`-th iteration.

    Returns
    -------
    pi_samp : (n_sample_x, n_sample_y) array-like, float
        Sample coupling matrix.
    pi_feat : (n_feature_x, n_feature_y) array-like, float
        Feature coupling matrix.
    log : dictionary, optional
        Returned if `log` is True. The keys are:

            error : array-like, float
                list of L1 norms between the current and previous sample coupling.
            duals_sample : (n_sample_x, n_sample_y) tuple, float
                Pair of dual vectors when solving OT problem w.r.t the sample coupling.
            duals_feature : (n_feature_x, n_feature_y) tuple, float
                Pair of dual vectors when solving OT problem w.r.t the feature coupling.
            linear : float
                Linear part of the cost.
            ucoot : float
                Total cost.
            backend
                The proper backend for all input arrays
    """

    # MAIN FUNCTION

    if reg_type not in ["joint", "independent"]:
        raise (NotImplementedError('Unknown reg_type="{}"'.format(reg_type)))
    if divergence not in ["kl", "l2"]:
        raise (NotImplementedError('Unknown divergence="{}"'.format(divergence)))
    if unbalanced_solver not in ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"]:
        raise (NotImplementedError('Unknown method="{}"'.format(unbalanced_solver)))

    # hyperparameters
    alpha_samp, alpha_feat = get_parameter_pair(alpha)
    rho_x, rho_y = get_parameter_pair(reg_marginals)
    eps_samp, eps_feat = get_parameter_pair(epsilon)

    if reg_type == "joint":  # same regularization
        eps_feat = eps_samp
    if unbalanced_solver in ["sinkhorn", "sinkhorn_log"] and divergence == "l2":
        warnings.warn(
            "Sinkhorn algorithm does not support L2 norm. \
                      Divergence is set to 'kl'."
        )
        divergence = "kl"
    if unbalanced_solver in ["sinkhorn", "sinkhorn_log"] and (
        eps_samp == 0 or eps_feat == 0
    ):
        warnings.warn(
            "Sinkhorn algorithm does not support unregularized problem. \
                      Solver is set to 'mm'."
        )
        unbalanced_solver = "mm"

    if init_pi is None:
        pi_samp, pi_feat = None, None
    else:
        pi_samp, pi_feat = init_pi

    if init_duals is None:
        init_duals = (None, None)
    duals_samp, duals_feat = init_duals

    arr = [X, Y]

    for tuple in [duals_samp, duals_feat]:
        if tuple is not None:
            d1, d2 = duals_feat
            if d1 is not None:
                arr.append(list_to_array(d1))
            if d2 is not None:
                arr.append(list_to_array(d2))

    nx = get_backend(
        *arr, wx_samp, wx_feat, wy_samp, wy_feat, M_samp, M_feat, pi_samp, pi_feat
    )

    # constant input variables
    if M_samp is None:
        if alpha_samp > 0:
            warnings.warn(
                "M_samp is None but alpha_samp = {} > 0. \
                          The algo will treat as if alpha_samp = 0.".format(alpha_samp)
            )
    else:
        M_samp = alpha_samp * M_samp

    if M_feat is None:
        if alpha_feat > 0:
            warnings.warn(
                "M_feat is None but alpha_feat = {} > 0. \
                          The algo will treat as if alpha_feat = 0.".format(alpha_feat)
            )
    else:
        M_feat = alpha_feat * M_feat

    nx_samp, nx_feat = X.shape
    ny_samp, ny_feat = Y.shape

    # measures on rows and columns
    if wx_samp is None:
        wx_samp = nx.ones(nx_samp, type_as=X) / nx_samp
    if wx_feat is None:
        wx_feat = nx.ones(nx_feat, type_as=X) / nx_feat
    if wy_samp is None:
        wy_samp = nx.ones(ny_samp, type_as=Y) / ny_samp
    if wy_feat is None:
        wy_feat = nx.ones(ny_feat, type_as=Y) / ny_feat
    wxy_samp = wx_samp[:, None] * wy_samp[None, :]
    wxy_feat = wx_feat[:, None] * wy_feat[None, :]

    # initialize coupling and dual vectors
    pi_samp = wxy_samp if pi_samp is None else pi_samp
    pi_feat = wxy_feat if pi_feat is None else pi_feat

    if unbalanced_solver in ["sinkhorn", "sinkhorn_log"]:
        if duals_samp is None:
            duals_samp = (nx.zeros(nx_samp, type_as=X), nx.zeros(ny_samp, type_as=Y))
        if duals_feat is None:
            duals_feat = (nx.zeros(nx_feat, type_as=X), nx.zeros(ny_feat, type_as=Y))

    # shortcut functions
    X_sqr, Y_sqr = X**2, Y**2
    local_cost_samp = partial(
        uot_cost_matrix,
        data=(X_sqr, Y_sqr, X, Y, M_samp),
        tuple_p=(wx_feat, wy_feat),
        hyperparams=(rho_x, rho_y, eps_feat),
        divergence=divergence,
        reg_type=reg_type,
        nx=nx,
    )

    local_cost_feat = partial(
        uot_cost_matrix,
        data=(X_sqr.T, Y_sqr.T, X.T, Y.T, M_feat),
        tuple_p=(wx_samp, wy_samp),
        hyperparams=(rho_x, rho_y, eps_samp),
        divergence=divergence,
        reg_type=reg_type,
        nx=nx,
    )

    parameters_uot_l2_samp = partial(
        uot_parameters_and_measures,
        tuple_weights=(wx_samp, wy_samp, wxy_samp),
        hyperparams=(rho_x, rho_y, eps_samp),
        reg_type=reg_type,
        divergence=divergence,
        nx=nx,
    )

    parameters_uot_l2_feat = partial(
        uot_parameters_and_measures,
        tuple_weights=(wx_feat, wy_feat, wxy_feat),
        hyperparams=(rho_x, rho_y, eps_feat),
        reg_type=reg_type,
        divergence=divergence,
        nx=nx,
    )

    solver = partial(
        ot.solve,
        reg_type=divergence,
        unbalanced_type=divergence,
        method=unbalanced_solver,
        max_iter=max_iter_ot,
        tol=tol_ot,
        verbose=False,
    )

    # initialize log
    if log:
        dict_log = {"backend": nx, "error": []}

    for idx in range(max_iter):
        pi_samp_prev = nx.copy(pi_samp)

        # Update feature coupling
        mass = nx.sum(pi_samp)
        uot_cost = local_cost_feat(pi=pi_samp)

        if divergence == "kl":
            new_rho = (rho_x * mass, rho_y * mass)
            new_eps = mass * eps_feat if reg_type == "joint" else eps_feat
            new_wx, new_wy, new_wxy = wx_feat, wy_feat, wxy_feat
        else:  # divergence == "l2"
            new_w, new_rho, new_eps = parameters_uot_l2_feat(pi_feat)
            new_wx, new_wy, new_wxy = new_w

        res = solver(
            M=uot_cost,
            a=new_wx,
            b=new_wy,
            reg=new_eps,
            c=new_wxy,
            unbalanced=new_rho,
            plan_init=pi_feat,
            potentials_init=duals_feat,
        )
        pi_feat, duals_feat = res.plan, res.potentials

        if rescale_plan:
            pi_feat = nx.sqrt(mass / nx.sum(pi_feat)) * pi_feat

        # Update sample coupling
        mass = nx.sum(pi_feat)
        uot_cost = local_cost_samp(pi=pi_feat)

        if divergence == "kl":
            new_rho = (rho_x * mass, rho_y * mass)
            new_eps = mass * eps_feat if reg_type == "joint" else eps_feat
            new_wx, new_wy, new_wxy = wx_samp, wy_samp, wxy_samp
        else:  # divergence == "l2"
            new_w, new_rho, new_eps = parameters_uot_l2_samp(pi_samp)
            new_wx, new_wy, new_wxy = new_w

        res = solver(
            M=uot_cost,
            a=new_wx,
            b=new_wy,
            reg=new_eps,
            c=new_wxy,
            unbalanced=new_rho,
            plan_init=pi_samp,
            potentials_init=duals_samp,
        )
        pi_samp, duals_samp = res.plan, res.potentials

        if rescale_plan:
            pi_samp = nx.sqrt(mass / nx.sum(pi_samp)) * pi_samp  # shape nx x ny

        # get L1 error
        err = nx.sum(nx.abs(pi_samp - pi_samp_prev))
        if log:
            dict_log["error"].append(err)
        if verbose:
            print("{:5d}|{:8e}|".format(idx + 1, err))
        if err < tol:
            break

    # sanity check
    if nx.sum(nx.isnan(pi_samp)) > 0 or nx.sum(nx.isnan(pi_feat)) > 0:
        raise (
            ValueError(
                "There is NaN in coupling. \
                          Adjust the relaxation or regularization parameters."
            )
        )

    if log:
        linear_cost, ucoot_cost = fused_unbalanced_across_spaces_cost(
            M_linear=(M_samp, M_feat),
            data=(X_sqr, Y_sqr, X, Y),
            tuple_pxy_samp=(wx_samp, wy_samp, wxy_samp),
            tuple_pxy_feat=(wx_feat, wy_feat, wxy_feat),
            pi_samp=pi_samp,
            pi_feat=pi_feat,
            hyperparams=(rho_x, rho_y, eps_samp, eps_feat),
            divergence=divergence,
            reg_type=reg_type,
            nx=nx,
        )

        dict_log["duals_sample"] = duals_samp
        dict_log["duals_feature"] = duals_feat
        dict_log["linear_cost"] = linear_cost
        dict_log["ucoot_cost"] = ucoot_cost

        return pi_samp, pi_feat, dict_log

    else:
        return pi_samp, pi_feat


def unbalanced_co_optimal_transport(
    X,
    Y,
    wx_samp=None,
    wx_feat=None,
    wy_samp=None,
    wy_feat=None,
    reg_marginals=10,
    epsilon=0,
    divergence="kl",
    unbalanced_solver="mm",
    alpha=0,
    M_samp=None,
    M_feat=None,
    rescale_plan=True,
    init_pi=None,
    init_duals=None,
    max_iter=100,
    tol=1e-7,
    max_iter_ot=500,
    tol_ot=1e-7,
    log=False,
    verbose=False,
    **kwargs_solve,
):
    r"""Compute the unbalanced Co-Optimal Transport between two Euclidean point clouds
    (represented as matrices whose rows are samples and columns are the features/dimensions).

    More precisely, this function returns the sample and feature transport plans between
    :math:`(\mathbf{X}, \mathbf{w}_{xs}, \mathbf{w}_{xf})` and
    :math:`(\mathbf{Y}, \mathbf{w}_{ys}, \mathbf{w}_{yf})`,
    by solving the following problem using Block Coordinate Descent algorithm:

    .. math::
        \mathop{\arg \min}_{\mathbf{P}, \mathbf{Q}} &\quad \sum_{i,j,k,l}
        (\mathbf{X}_{i,k} - \mathbf{Y}_{j,l})^2 \mathbf{P}_{i,j} \mathbf{Q}_{k,l} \\
        &+ \rho_s \mathbf{Div}(\mathbf{P}_{\# 1} \mathbf{Q}_{\# 1}^T | \mathbf{w}_{xs} \mathbf{w}_{ys}^T)
        + \rho_f \mathbf{Div}(\mathbf{P}_{\# 2} \mathbf{Q}_{\# 2}^T | \mathbf{w}_{xf} \mathbf{w}_{yf}^T) \\
        &+ \alpha_s \sum_{i,j} \mathbf{P}_{i,j} \mathbf{M^{(s)}}_{i, j}
        + \alpha_f \sum_{k, l} \mathbf{Q}_{k,l} \mathbf{M^{(f)}}_{k, l} \\
        &+ \varepsilon_s \mathbf{Div}(\mathbf{P} | \mathbf{w}_{xs} \mathbf{w}_{ys}^T)
        + \varepsilon_f \mathbf{Div}(\mathbf{Q} | \mathbf{w}_{xf} \mathbf{w}_{yf}^T)

    Where:

    - :math:`\mathbf{X}`: Source input (arbitrary-size) matrix
    - :math:`\mathbf{Y}`: Target input (arbitrary-size) matrix
    - :math:`\mathbf{M^{(s)}}`: Additional sample matrix
    - :math:`\mathbf{M^{(f)}}`: Additional feature matrix
    - :math:`\mathbf{w}_{xs}`: Distribution of the samples in the source space
    - :math:`\mathbf{w}_{xf}`: Distribution of the features in the source space
    - :math:`\mathbf{w}_{ys}`: Distribution of the samples in the target space
    - :math:`\mathbf{w}_{yf}`: Distribution of the features in the target space
    - :math:`\mathbf{Div}`: Either Kullback-Leibler divergence or half-squared L2 norm.

    .. note:: This function allows `epsilon` to be zero. In that case, `unbalanced_method` must be either "mm" or "lbfgsb".

    Parameters
    ----------
    X : (n_sample_x, n_feature_x) array-like, float
        Source input matrix.
    Y : (n_sample_y, n_feature_y) array-like, float
        Target input matrix.
    wx_samp : (n_sample_x, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix X.
        Uniform distribution by default.
    wx_feat : (n_feature_x, ) array-like, float, optional (default = None)
        Histogram assigned on columns (features) of matrix X.
        Uniform distribution by default.
    wy_samp : (n_sample_y, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix Y.
        Uniform distribution by default.
    wy_feat : (n_feature_y, ) array-like, float, optional (default = None)
        Histogram assigned on columns (features) of matrix Y.
        Uniform distribution by default.
    reg_marginals: float or indexable object of length 1 or 2
        Marginal relaxation terms for sample and feature couplings.
        If `reg_marginals is a scalar` or an indexable object of length 1,
        then the same value is applied to both marginal relaxations.
    epsilon : scalar or indexable object of length 2, float or int, optional (default = 0)
        Regularization parameters for entropic approximation of sample and feature couplings.
        Allow the case where `epsilon` contains 0. In that case, the MM solver is used by default
        instead of Sinkhorn solver. If `epsilon` is scalar, then the same value is applied to
        both regularization of sample and feature couplings.
    divergence : string, optional (default = "kl")

        - If `divergence` = "kl", then Div is the Kullback-Leibler divergence.

        - If `divergence` = "l2", then Div is the half squared Euclidean norm.
    unbalanced_solver : string, optional (default = "sinkhorn")
        Solver for the unbalanced OT subroutine.

        - If `divergence` = "kl", then `unbalanced_solver` can be: "sinkhorn", "sinkhorn_log", "mm", "lbfgsb"

        - If `divergence` = "l2", then `unbalanced_solver` can be "mm", "lbfgsb"
    alpha : scalar or indexable object of length 2, float or int, optional (default = 0)
        Coeffficient parameter of linear terms with respect to the sample and feature couplings.
        If alpha is scalar, then the same alpha is applied to both linear terms.
    M_samp : (n_sample_x, n_sample_y), float, optional (default = None)
        Sample matrix associated to the Wasserstein linear term on sample coupling.
    M_feat : (n_feature_x, n_feature_y), float, optional (default = None)
        Feature matrix associated to the Wasserstein linear term on feature coupling.
    rescale_plan : boolean, optional (default = True)
        If True, then rescale the sample and feature transport plans within each BCD iteration,
        so that they always have equal mass.
    init_pi : tuple of two matrices of size (n_sample_x, n_sample_y) and
        (n_feature_x, n_feature_y), optional (default = None).
        Initialization of sample and feature couplings.
        Uniform distributions by default.
    init_duals : tuple of two tuples ((n_sample_x, ), (n_sample_y, )) and ((n_feature_x, ), (n_feature_y, )), optional (default = None).
        Initialization of sample and feature dual vectors
        if using Sinkhorn algorithm. Zero vectors by default.
    max_iter : int, optional (default = 100)
        Number of Block Coordinate Descent (BCD) iterations.
    tol : float, optional (default = 1e-7)
        Tolerance of BCD scheme. If the L1-norm between the current and previous
        sample couplings is under this threshold, then stop BCD scheme.
    max_iter_ot : int, optional (default = 100)
        Number of iterations to solve each of the
        two unbalanced optimal transport problems in each BCD iteration.
    tol_ot : float, optional (default = 1e-7)
        Tolerance of unbalanced solver for each of the
        two unbalanced optimal transport problems in each BCD iteration.
    log : bool, optional (default = False)
        If True then the cost and four dual vectors, including
        two from sample and two from feature couplings, are recorded.
    verbose : bool, optional (default = False)
        If True then print the COOT cost at every multiplier of `eval_bcd`-th iteration.

    Returns
    -------
    pi_samp : (n_sample_x, n_sample_y) array-like, float
        Sample coupling matrix.
    pi_feat : (n_feature_x, n_feature_y) array-like, float
        Feature coupling matrix.
    log : dictionary, optional
        Returned if `log` is True. The keys are:

            error : array-like, float
                list of L1 norms between the current and previous sample coupling.
            duals_sample : (n_sample_x, n_sample_y)-tuple, float
                Pair of dual vectors when solving OT problem w.r.t the sample coupling.
            duals_feature : (n_feature_x, n_feature_y)-tuple, float
                Pair of dual vectors when solving OT problem w.r.t the feature coupling.
            linear : float
                Linear part of the cost.
            ucoot : float
                Total cost.

    References
    ----------
    .. [71] Tran, H., Janati, H., Courty, N., Flamary, R., Redko, I., Demetci, P., & Singh, R.
            Unbalanced Co-Optimal Transport. AAAI Conference on Artificial Intelligence, 2023.
    """

    return fused_unbalanced_across_spaces_divergence(
        X=X,
        Y=Y,
        wx_samp=wx_samp,
        wx_feat=wx_feat,
        wy_samp=wy_samp,
        wy_feat=wy_feat,
        reg_marginals=reg_marginals,
        epsilon=epsilon,
        reg_type="independent",
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp,
        M_feat=M_feat,
        rescale_plan=rescale_plan,
        init_pi=init_pi,
        init_duals=init_duals,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=log,
        verbose=verbose,
        **kwargs_solve,
    )


def unbalanced_co_optimal_transport2(
    X,
    Y,
    wx_samp=None,
    wx_feat=None,
    wy_samp=None,
    wy_feat=None,
    reg_marginals=10,
    epsilon=0,
    divergence="kl",
    unbalanced_solver="sinkhorn",
    alpha=0,
    M_samp=None,
    M_feat=None,
    rescale_plan=True,
    init_pi=None,
    init_duals=None,
    max_iter=100,
    tol=1e-7,
    max_iter_ot=500,
    tol_ot=1e-7,
    log=False,
    verbose=False,
    **kwargs_solve,
):
    r"""Compute the unbalanced Co-Optimal Transport between two Euclidean point clouds
    (represented as matrices whose rows are samples and columns are the features/dimensions).

    More precisely, this function returns the unbalanced Co-Optimal Transport cost between
    :math:`(\mathbf{X}, \mathbf{w}_{xs}, \mathbf{w}_{xf})` and
    :math:`(\mathbf{Y}, \mathbf{w}_{ys}, \mathbf{w}_{yf})`,
    by solving the following problem using Block Coordinate Descent algorithm:

    .. math::
        \mathop{\min}_{\mathbf{P}, \mathbf{Q}} &\quad \sum_{i,j,k,l}
        (\mathbf{X}_{i,k} - \mathbf{Y}_{j,l})^2 \mathbf{P}_{i,j} \mathbf{Q}_{k,l} \\
        &+ \rho_s \mathbf{Div}(\mathbf{P}_{\# 1} \mathbf{Q}_{\# 1}^T | \mathbf{w}_{xs} \mathbf{w}_{ys}^T)
        + \rho_f \mathbf{Div}(\mathbf{P}_{\# 2} \mathbf{Q}_{\# 2}^T | \mathbf{w}_{xf} \mathbf{w}_{yf}^T) \\
        &+ \alpha_s \sum_{i,j} \mathbf{P}_{i,j} \mathbf{M^{(s)}}_{i, j}
        + \alpha_f \sum_{k, l} \mathbf{Q}_{k,l} \mathbf{M^{(f)}}_{k, l} \\
        &+ \varepsilon_s \mathbf{Div}(\mathbf{P} | \mathbf{w}_{xs} \mathbf{w}_{ys}^T)
        + \varepsilon_f \mathbf{Div}(\mathbf{Q} | \mathbf{w}_{xf} \mathbf{w}_{yf}^T)

    Where:

    - :math:`\mathbf{X}`: Source input (arbitrary-size) matrix
    - :math:`\mathbf{Y}`: Target input (arbitrary-size) matrix
    - :math:`\mathbf{M^{(s)}}`: Additional sample matrix
    - :math:`\mathbf{M^{(f)}}`: Additional feature matrix
    - :math:`\mathbf{w}_{xs}`: Distribution of the samples in the source space
    - :math:`\mathbf{w}_{xf}`: Distribution of the features in the source space
    - :math:`\mathbf{w}_{ys}`: Distribution of the samples in the target space
    - :math:`\mathbf{w}_{yf}`: Distribution of the features in the target space
    - :math:`\mathbf{Div}`: Either Kullback-Leibler divergence or half-squared L2 norm.

    .. note:: This function allows `epsilon` to be zero. In that case, `unbalanced_method` must be either "mm" or "lbfgsb".
            Also the computation of gradients is only supported for KL divergence. The case of half squared-L2 norm uses those of KL divergence.

    Parameters
    ----------
    X : (n_sample_x, n_feature_x) array-like, float
        Source input matrix.
    Y : (n_sample_y, n_feature_y) array-like, float
        Target input matrix.
    wx_samp : (n_sample_x, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix X.
        Uniform distribution by default.
    wx_feat : (n_feature_x, ) array-like, float, optional (default = None)
        Histogram assigned on columns (features) of matrix X.
        Uniform distribution by default.
    wy_samp : (n_sample_y, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix Y.
        Uniform distribution by default.
    wy_feat : (n_feature_y, ) array-like, float, optional (default = None)
        Histogram assigned on columns (features) of matrix Y.
        Uniform distribution by default.
    reg_marginals: float or indexable object of length 1 or 2
        Marginal relaxation terms for sample and feature couplings.
        If `reg_marginals` is a scalar or an indexable object of length 1,
        then the same value is applied to both marginal relaxations.
    epsilon : scalar or indexable object of length 2, float or int, optional (default = 0)
        Regularization parameters for entropic approximation of sample and feature couplings.
        Allow the case where `epsilon` contains 0. In that case, the MM solver is used by default
        instead of Sinkhorn solver. If `epsilon` is scalar, then the same value is applied to
        both regularization of sample and feature couplings.
    divergence : string, optional (default = "kl")

        - If `divergence` = "kl", then Div is the Kullback-Leibler divergence.

        - If `divergence` = "l2", then Div is the half squared Euclidean norm.
    unbalanced_solver : string, optional (default = "sinkhorn")
        Solver for the unbalanced OT subroutine.

        - If `divergence` = "kl", then `unbalanced_solver` can be: "sinkhorn", "sinkhorn_log", "mm", "lbfgsb"

        - If `divergence` = "l2", then `unbalanced_solver` can be "mm", "lbfgsb"
    alpha : scalar or indexable object of length 2, float or int, optional (default = 0)
        Coeffficient parameter of linear terms with respect to the sample and feature couplings.
        If alpha is scalar, then the same alpha is applied to both linear terms.
    M_samp : (n_sample_x, n_sample_y), float, optional (default = None)
        Sample matrix associated to the Wasserstein linear term on sample coupling.
    M_feat : (n_feature_x, n_feature_y), float, optional (default = None)
        Feature matrix associated to the Wasserstein linear term on feature coupling.
    rescale_plan : boolean, optional (default = True)
        If True, then rescale the transport plans in each BCD iteration,
        so that they always have equal mass.
    init_pi : tuple of two matrices of size (n_sample_x, n_sample_y) and
        (n_feature_x, n_feature_y), optional (default = None).
        Initialization of sample and feature couplings.
        Uniform distributions by default.
    init_duals : tuple of two tuples ((n_sample_x, ), (n_sample_y, )) and ((n_feature_x, ), (n_feature_y, )), optional (default = None).
        Initialization of sample and feature dual vectors
        if using Sinkhorn algorithm. Zero vectors by default.
    max_iter : int, optional (default = 100)
        Number of Block Coordinate Descent (BCD) iterations.
    tol : float, optional (default = 1e-7)
        Tolerance of BCD scheme. If the L1-norm between the current and previous
        sample couplings is under this threshold, then stop BCD scheme.
    max_iter_ot : int, optional (default = 100)
        Number of iterations to solve each of the
        two unbalanced optimal transport problems in each BCD iteration.
    tol_ot : float, optional (default = 1e-7)
        Tolerance of unbalanced solver for each of the
        two unbalanced optimal transport problems in each BCD iteration.
    log : bool, optional (default = False)
        If True then the cost and four dual vectors, including
        two from sample and two from feature couplings, are recorded.
    verbose : bool, optional (default = False)
        If True then print the COOT cost at every multiplier of `eval_bcd`-th iteration.

    Returns
    -------
    ucoot : float
        UCOOT cost.
    log : dictionary, optional
        Returned if `log` is True. The keys are:

            error : array-like, float
                list of L1 norms between the current and previous sample coupling.
            duals_sample : (n_sample_x, n_sample_y)-tuple, float
                Pair of dual vectors when solving OT problem w.r.t the sample coupling.
            duals_feature : (n_feature_x, n_feature_y)-tuple, float
                Pair of dual vectors when solving OT problem w.r.t the feature coupling.
            linear : float
                Linear part of UCOOT cost.
            ucoot : float
                UCOOT cost.
            backend
                The proper backend for all input arrays

    References
    ----------
    .. [71] Tran, H., Janati, H., Courty, N., Flamary, R., Redko, I., Demetci, P., & Singh, R.
            Unbalanced Co-Optimal Transport. AAAI Conference on Artificial Intelligence, 2023.
    """

    if divergence != "kl":
        warnings.warn(
            "The computation of gradients is only supported for KL divergence, not \
                      for {} divergence".format(divergence)
        )

    pi_samp, pi_feat, log_ucoot = unbalanced_co_optimal_transport(
        X=X,
        Y=Y,
        wx_samp=wx_samp,
        wx_feat=wx_feat,
        wy_samp=wy_samp,
        wy_feat=wy_feat,
        reg_marginals=reg_marginals,
        epsilon=epsilon,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp,
        M_feat=M_feat,
        rescale_plan=rescale_plan,
        init_pi=init_pi,
        init_duals=init_duals,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=True,
        verbose=verbose,
        **kwargs_solve,
    )

    nx = log_ucoot["backend"]

    nx_samp, nx_feat = X.shape
    ny_samp, ny_feat = Y.shape

    # measures on rows and columns
    if wx_samp is None:
        wx_samp = nx.ones(nx_samp, type_as=X) / nx_samp
    if wx_feat is None:
        wx_feat = nx.ones(nx_feat, type_as=X) / nx_feat
    if wy_samp is None:
        wy_samp = nx.ones(ny_samp, type_as=Y) / ny_samp
    if wy_feat is None:
        wy_feat = nx.ones(ny_feat, type_as=Y) / ny_feat

    # extract parameters
    rho_x, rho_y = get_parameter_pair(reg_marginals)
    eps_samp, eps_feat = get_parameter_pair(epsilon)

    # calculate marginals
    pi1_samp, pi2_samp = nx.sum(pi_samp, 1), nx.sum(pi_samp, 0)
    pi1_feat, pi2_feat = nx.sum(pi_feat, 1), nx.sum(pi_feat, 0)
    m_samp, m_feat = nx.sum(pi1_samp), nx.sum(pi1_feat)
    m_wx_feat, m_wx_samp = nx.sum(wx_feat), nx.sum(wx_samp)
    m_wy_feat, m_wy_samp = nx.sum(wy_feat), nx.sum(wy_samp)

    # calculate subgradients
    gradX = 2 * X * (pi1_samp[:, None] * pi1_feat[None, :]) - 2 * nx.dot(
        nx.dot(pi_samp, Y), pi_feat.T
    )  # shape (nx_samp, nx_feat)
    gradY = 2 * Y * (pi2_samp[:, None] * pi2_feat[None, :]) - 2 * nx.dot(
        nx.dot(pi_samp.T, X), pi_feat
    )  # shape (ny_samp, ny_feat)

    grad_wx_samp = rho_x * (m_wx_feat - m_feat * pi1_samp / wx_samp) + eps_samp * (
        m_wy_samp - pi1_samp / wx_samp
    )
    grad_wx_feat = rho_x * (m_wx_samp - m_samp * pi1_feat / wx_feat) + eps_feat * (
        m_wy_feat - pi1_feat / wx_feat
    )
    grad_wy_samp = rho_y * (m_wy_feat - m_feat * pi2_samp / wy_samp) + eps_samp * (
        m_wx_samp - pi2_samp / wy_samp
    )
    grad_wy_feat = rho_y * (m_wy_samp - m_samp * pi2_feat / wy_feat) + eps_feat * (
        m_wx_feat - pi2_feat / wy_feat
    )

    # set gradients
    ucoot = log_ucoot["ucoot_cost"]
    ucoot = nx.set_gradients(
        ucoot,
        (X, Y, wx_samp, wx_feat, wy_samp, wy_feat),
        (gradX, gradY, grad_wx_samp, grad_wx_feat, grad_wy_samp, grad_wy_feat),
    )

    if log:
        return ucoot, log_ucoot

    else:
        return ucoot


def fused_unbalanced_gromov_wasserstein(
    Cx,
    Cy,
    wx=None,
    wy=None,
    reg_marginals=10,
    epsilon=0,
    divergence="kl",
    unbalanced_solver="mm",
    alpha=0,
    M=None,
    init_duals=None,
    init_pi=None,
    max_iter=100,
    tol=1e-7,
    max_iter_ot=500,
    tol_ot=1e-7,
    log=False,
    verbose=False,
    **kwargs_solve,
):
    r"""Compute the lower bound of the fused unbalanced Gromov-Wasserstein (FUGW) between two similarity matrices.
    In practice, this lower bound is used interchangeably with the true FUGW.

    More precisely, this function returns the transport plan between
    :math:`(\mathbf{C^X}, \mathbf{w_X})` and :math:`(\mathbf{C^Y}, \mathbf{w_Y})`,
    by solving the following problem using Block Coordinate Descent algorithm:

    .. math::
        \mathop{\arg \min}_{\substack{\mathbf{P}, \mathbf{Q}: \\ mass(P) = mass(Q)}}
        &\quad \sum_{i,j,k,l} (\mathbf{C^X}_{i,k} - \mathbf{C^Y}_{j,l})^2 \mathbf{P}_{i,j} \mathbf{Q}_{k,l}
        + \frac{\alpha}{2} \sum_{i,j} (\mathbf{P}_{i,j} + \mathbf{Q}_{i,j}) \mathbf{M}_{i, j} \\
        &+ \rho_1 \mathbf{Div}(\mathbf{P}_{\# 1} \mathbf{Q}_{\# 1}^T | \mathbf{w_X} \mathbf{w_X}^T)
        + \rho_2 \mathbf{Div}(\mathbf{P}_{\# 2} \mathbf{Q}_{\# 2}^T | \mathbf{w_Y} \mathbf{w_Y}^T) \\
        &+ \varepsilon \mathbf{Div}(\mathbf{P} \otimes \mathbf{Q} | (\mathbf{w_X} \mathbf{w_Y}^T) \otimes (\mathbf{w_X} \mathbf{w_Y}^T))

    Where:

    - :math:`\mathbf{C^X}`: Source similarity matrix
    - :math:`\mathbf{C^Y}`: Target similarity matrix
    - :math:`\mathbf{M}`: Sample matrix corresponding to the Wasserstein term
    - :math:`\mathbf{w_X}`: Distribution of the samples in the source space
    - :math:`\mathbf{w_Y}`: Distribution of the samples in the target space
    - :math:`\mathbf{Div}`: Either Kullback-Leibler divergence or half-squared L2 norm.

    .. note:: This function allows epsilon to be zero. In that case, `unbalanced_method` must be either "mm" or "lbfgsb".

    Parameters
    ----------
    Cx : (n_sample_x, n_feature_x) array-like, float
        Source similarity matrix.
    Cy : (n_sample_y, n_feature_y) array-like, float
        Target similarity matrix.
    wx : (n_sample_x, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix Cx.
        Uniform distribution by default.
    wy : (n_sample_y, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix Cy.
        Uniform distribution by default.
    reg_marginals: float or indexable object of length 1 or 2
        Marginal relaxation terms for sample and feature couplings.
        If `reg_marginals` is a scalar or an indexable object of length 1,
        then the same value is applied to both marginal relaxations.
    epsilon : scalar, float or int, optional (default = 0)
        Regularization parameters for entropic approximation of sample and feature couplings.
        Allow the case where `epsilon` contains 0. In that case, the MM solver is used by default
        instead of Sinkhorn solver. If `epsilon` is scalar, then the same value is applied to
        both regularization of sample and feature couplings.
    divergence : string, optional (default = "kl")

        - If `divergence` = "kl", then Div is the Kullback-Leibler divergence.

        - If `divergence` = "l2", then Div is the half squared Euclidean norm.
    unbalanced_solver : string, optional (default = "sinkhorn")
        Solver for the unbalanced OT subroutine.

        - If `divergence` = "kl", then `unbalanced_solver` can be: "sinkhorn", "sinkhorn_log", "mm", "lbfgsb"

        - If `divergence` = "l2", then `unbalanced_solver` can be "mm", "lbfgsb"
    alpha : scalar, float or int, optional (default = 0)
        Coeffficient parameter of linear terms with respect to the sample and feature couplings.
        If alpha is scalar, then the same alpha is applied to both linear terms.
    M : (n_sample_x, n_sample_y), float, optional (default = None)
        Sample matrix associated to the Wasserstein linear term on sample coupling.
    init_pi :(n_sample_x, n_sample_y) array-like, optional (default = None)
        Initialization of sample coupling. By default = :math:`w_X w_Y^T`.
    init_duals : tuple of vectors ((n_sample_x, ), (n_sample_y, )), optional (default = None).
        Initialization of sample and feature dual vectors
        if using Sinkhorn algorithm. Zero vectors by default.
    max_iter : int, optional (default = 100)
        Number of Block Coordinate Descent (BCD) iterations.
    tol : float, optional (default = 1e-7)
        Tolerance of BCD scheme. If the L1-norm between the current and previous
        sample couplings is under this threshold, then stop BCD scheme.
    max_iter_ot : int, optional (default = 100)
        Number of iterations to solve each of the
        two unbalanced optimal transport problems in each BCD iteration.
    tol_ot : float, optional (default = 1e-7)
        Tolerance of unbalanced solver for each of the
        two unbalanced optimal transport problems in each BCD iteration.
    log : bool, optional (default = False)
        If True then the cost and four dual vectors, including
        two from sample and two from feature couplings, are recorded.
    verbose : bool, optional (default = False)
        If True then print the COOT cost at every multiplier of `eval_bcd`-th iteration.

    Returns
    -------
    pi_samp : (n_sample_x, n_sample_y) array-like, float
        Sample coupling matrix.
        In practice, we use this matrix as solution of FUGW.
    pi_samp2 : (n_sample_x, n_sample_y) array-like, float
        Second sample coupling matrix.
        In practice, we usually ignore this output.
    log : dictionary, optional
        Returned if `log` is True. The keys are:

            error : array-like, float
                list of L1 norms between the current and previous sample couplings.
            duals : (n_sample_x, n_sample_y)-tuple, float
                Pair of dual vectors when solving OT problem w.r.t the sample coupling.
            linear : float
                Linear part of FUGW cost.
            fugw_cost : float
                Total FUGW cost.
            backend
                The proper backend for all input arrays

    References
    ----------
    .. [70] Thual, A., Tran, H., Zemskova, T., Courty, N., Flamary, R., Dehaene, S., & Thirion, B.
            Aligning individual brains with Fused Unbalanced Gromov-Wasserstein.
            Advances in Neural Information Systems, 35 (2022).

    .. [72] Thibault Séjourné, François-Xavier Vialard, & Gabriel Peyré.
            The Unbalanced Gromov Wasserstein Distance: Conic Formulation and Relaxation.
            Neural Information Processing Systems, 34 (2021).
    """

    alpha = (alpha / 2, alpha / 2)

    pi_samp, pi_feat, dict_log = fused_unbalanced_across_spaces_divergence(
        X=Cx,
        Y=Cy,
        wx_samp=wx,
        wx_feat=wx,
        wy_samp=wy,
        wy_feat=wy,
        reg_marginals=reg_marginals,
        epsilon=epsilon,
        reg_type="joint",
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M,
        M_feat=M,
        rescale_plan=True,
        init_pi=(init_pi, init_pi),
        init_duals=(init_duals, init_duals),
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=True,
        verbose=verbose,
        **kwargs_solve,
    )

    if log:
        log_fugw = {
            "error": dict_log["error"],
            "duals": dict_log["duals_sample"],
            "linear_cost": dict_log["linear_cost"],
            "fugw_cost": dict_log["ucoot_cost"],
            "backend": dict_log["backend"],
        }

        return pi_samp, pi_feat, log_fugw

    else:
        return pi_samp, pi_feat


def fused_unbalanced_gromov_wasserstein2(
    Cx,
    Cy,
    wx=None,
    wy=None,
    reg_marginals=10,
    epsilon=0,
    divergence="kl",
    unbalanced_solver="mm",
    alpha=0,
    M=None,
    init_duals=None,
    init_pi=None,
    max_iter=100,
    tol=1e-7,
    max_iter_ot=500,
    tol_ot=1e-7,
    log=False,
    verbose=False,
    **kwargs_solve,
):
    r"""Compute the lower bound of the fused unbalanced Gromov-Wasserstein (FUGW) between two similarity matrices.
    In practice, this lower bound is used interchangeably with the true FUGW.

    More precisely, this function returns the lower bound of the fused unbalanced Gromov-Wasserstein cost between
    :math:`(\mathbf{C^X}, \mathbf{w_X})` and :math:`(\mathbf{C^Y}, \mathbf{w_Y})`,
    by solving the following problem using Block Coordinate Descent algorithm:

    .. math::
        \mathop{\min}_{\substack{\mathbf{P}, \mathbf{Q}: \\ mass(P) = mass(Q)}}
        &\quad \sum_{i,j,k,l} (\mathbf{C^X}_{i,k} - \mathbf{C^Y}_{j,l})^2 \mathbf{P}_{i,j} \mathbf{Q}_{k,l}
        + \frac{\alpha}{2} \sum_{i,j} (\mathbf{P}_{i,j} + \mathbf{Q}_{i,j}) \mathbf{M}_{i, j} \\
        &+ \rho_1 \mathbf{Div}(\mathbf{P}_{\# 1} \mathbf{Q}_{\# 1}^T | \mathbf{w_X} \mathbf{w_X}^T)
        + \rho_2 \mathbf{Div}(\mathbf{P}_{\# 2} \mathbf{Q}_{\# 2}^T | \mathbf{w_Y} \mathbf{w_Y}^T) \\
        &+ \varepsilon \mathbf{Div}(\mathbf{P} \otimes \mathbf{Q} | (\mathbf{w_X} \mathbf{w_Y}^T) \otimes (\mathbf{w_X} \mathbf{w_Y}^T))

    Where:

    - :math:`\mathbf{C^X}`: Source similarity matrix
    - :math:`\mathbf{C^Y}`: Target similarity matrix
    - :math:`\mathbf{M}`: Sample matrix corresponding to the Wasserstein term
    - :math:`\mathbf{w_X}`: Distribution of the samples in the source space
    - :math:`\mathbf{w_Y}`: Distribution of the samples in the target space
    - :math:`\mathbf{Div}`: Either Kullback-Leibler divergence or half-squared L2 norm.

    .. note:: This function allows `epsilon` to be zero. In that case, unbalanced_method must be either "mm" or "lbfgsb".
            Also the computation of gradients is only supported for KL divergence, but not for half squared-L2 norm. In case of half squared-L2 norm, the calculation of KL divergence will be used.

    Parameters
    ----------
    Cx : (n_sample_x, n_feature_x) array-like, float
        Source similarity matrix.
    Cy : (n_sample_y, n_feature_y) array-like, float
        Target similarity matrix.
    wx : (n_sample_x, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix Cx.
        Uniform distribution by default.
    wy : (n_sample_y, ) array-like, float, optional (default = None)
        Histogram assigned on rows (samples) of matrix Cy.
        Uniform distribution by default.
    reg_marginals: float or indexable object of length 1 or 2
        Marginal relaxation terms for sample and feature couplings.
        If `reg_marginals` is a scalar or an indexable object of length 1,
        then the same value is applied to both marginal relaxations.
    epsilon : scalar, float or int, optional (default = 0)
        Regularization parameters for entropic approximation of sample and feature couplings.
        Allow the case where `epsilon` contains 0. In that case, the MM solver is used by default
        instead of Sinkhorn solver. If `epsilon` is scalar, then the same value is applied to
        both regularization of sample and feature couplings.
    divergence : string, optional (default = "kl")

        - If `divergence` = "kl", then Div is the Kullback-Leibler divergence.

        - If `divergence` = "l2", then Div is the half squared Euclidean norm.
    unbalanced_solver : string, optional (default = "sinkhorn")
        Solver for the unbalanced OT subroutine.

        - If `divergence` = "kl", then `unbalanced_solver` can be: "sinkhorn", "sinkhorn_log", "mm", "lbfgsb"

        - If `divergence` = "l2", then `unbalanced_solver` can be "mm", "lbfgsb"
    alpha : scalar, float or int, optional (default = 0)
        Coeffficient parameter of linear terms with respect to the sample and feature couplings.
        If alpha is scalar, then the same alpha is applied to both linear terms.
    M : (n_sample_x, n_sample_y), float, optional (default = None)
        Sample matrix associated to the Wasserstein linear term on sample coupling.
    init_pi :(n_sample_x, n_sample_y) array-like, optional (default = None)
        Initialization of sample coupling. By default = :math:`w_X w_Y^T`.
    init_duals : tuple of vectors ((n_sample_x, ), (n_sample_y, )), optional (default = None).
        Initialization of sample and feature dual vectors
        if using Sinkhorn algorithm. Zero vectors by default.
    max_iter : int, optional (default = 100)
        Number of Block Coordinate Descent (BCD) iterations.
    tol : float, optional (default = 1e-7)
        Tolerance of BCD scheme. If the L1-norm between the current and previous
        sample couplings is under this threshold, then stop BCD scheme.
    max_iter_ot : int, optional (default = 100)
        Number of iterations to solve each of the
        two unbalanced optimal transport problems in each BCD iteration.
    tol_ot : float, optional (default = 1e-7)
        Tolerance of unbalanced solver for each of the
        two unbalanced optimal transport problems in each BCD iteration.
    log : bool, optional (default = False)
        If True then the cost and four dual vectors, including
        two from sample and two from feature couplings, are recorded.
    verbose : bool, optional (default = False)
        If True then print the COOT cost at every multiplier of `eval_bcd`-th iteration.

    Returns
    -------
    fugw : float
        Total FUGW cost
    log : dictionary, optional
        Returned if `log` is True. The keys are:

            error : array-like, float
                list of L1 norms between the current and previous sample couplings.
            duals : (n_sample_x, n_sample_y)-tuple, float
                Pair of dual vectors when solving OT problem w.r.t the sample coupling.
            linear : float
                Linear part of FUGW cost.
            fugw_cost : float
                Total FUGW cost.
            backend
                The proper backend for all input arrays

    References
    ----------
    .. [70] Thual, A., Tran, H., Zemskova, T., Courty, N., Flamary, R., Dehaene, S., & Thirion, B.
            Aligning individual brains with Fused Unbalanced Gromov-Wasserstein.
            Advances in Neural Information Systems, 35 (2022).

    .. [72] Thibault Séjourné, François-Xavier Vialard, & Gabriel Peyré.
            The Unbalanced Gromov Wasserstein Distance: Conic Formulation and Relaxation.
            Neural Information Processing Systems, 34 (2021).
    """

    if divergence != "kl":
        warnings.warn(
            "The computation of gradients is only supported for KL divergence, \
                      but not for {} divergence. The gradient of the KL case will be used.".format(
                divergence
            )
        )

    pi_samp, pi_feat, log_fugw = fused_unbalanced_gromov_wasserstein(
        Cx=Cx,
        Cy=Cy,
        wx=wx,
        wy=wy,
        reg_marginals=reg_marginals,
        epsilon=epsilon,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M=M,
        init_duals=init_duals,
        init_pi=init_pi,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=True,
        verbose=verbose,
        **kwargs_solve,
    )

    nx = log_fugw["backend"]
    sx, sy = Cx.shape[0], Cy.shape[0]

    # measures on rows and columns
    if wx is None:
        wx = nx.ones(sx, type_as=Cx) / sx
    if wy is None:
        wy = nx.ones(sy, type_as=Cy) / sy

    # calculate marginals
    pi1_samp, pi2_samp = nx.sum(pi_samp, 1), nx.sum(pi_samp, 0)
    pi1_feat, pi2_feat = nx.sum(pi_feat, 1), nx.sum(pi_feat, 0)
    m_samp, m_feat = nx.sum(pi1_samp), nx.sum(pi1_feat)
    m_wx, m_wy = nx.sum(wx), nx.sum(wy)

    # calculate subgradients
    gradX = 2 * Cx * (pi1_samp[:, None] * pi1_feat[None, :]) - 2 * nx.dot(
        nx.dot(pi_samp, Cy), pi_feat.T
    )  # shape (nx_samp, nx_feat)
    gradY = 2 * Cy * (pi2_samp[:, None] * pi2_feat[None, :]) - 2 * nx.dot(
        nx.dot(pi_samp.T, Cx), pi_feat
    )  # shape (ny_samp, ny_feat)

    gradM = alpha / 2 * (pi_samp + pi_feat)

    rho_x, rho_y = get_parameter_pair(reg_marginals)
    grad_wx = (
        2 * m_wx * (rho_x + epsilon * m_wy**2)
        - (rho_x + epsilon) * (m_feat * pi1_samp + m_samp * pi1_feat) / wx
    )
    grad_wy = (
        2 * m_wy * (rho_y + epsilon * m_wx**2)
        - (rho_y + epsilon) * (m_feat * pi2_samp + m_samp * pi2_feat) / wy
    )

    # set gradients
    fugw = log_fugw["fugw_cost"]
    fugw = nx.set_gradients(
        fugw, (Cx, Cy, M, wx, wy), (gradX, gradY, gradM, grad_wx, grad_wy)
    )

    if log:
        return fugw, log_fugw

    else:
        return fugw



# -*- coding: utf-8 -*-
"""
Solvers related to Unbalanced Optimal Transport problems.

"""

# Author: Quang Huy Tran <quang-huy.tran@univ-ubs.fr>
#
# License: MIT License

# All submodules and packages
from ._sinkhorn import (
    sinkhorn_knopp_unbalanced,
    sinkhorn_unbalanced,
    sinkhorn_stabilized_unbalanced,
    sinkhorn_unbalanced_translation_invariant,
    sinkhorn_unbalanced2,
    barycenter_unbalanced_sinkhorn,
    barycenter_unbalanced_stabilized,
    barycenter_unbalanced,
)

from ._mm import mm_unbalanced, mm_unbalanced2

from ._lbfgs import lbfgsb_unbalanced, lbfgsb_unbalanced2

__all__ = [
    "sinkhorn_knopp_unbalanced",
    "sinkhorn_unbalanced",
    "sinkhorn_stabilized_unbalanced",
    "sinkhorn_unbalanced_translation_invariant",
    "sinkhorn_unbalanced2",
    "barycenter_unbalanced_sinkhorn",
    "barycenter_unbalanced_stabilized",
    "barycenter_unbalanced",
    "mm_unbalanced",
    "mm_unbalanced2",
    "_get_loss_unbalanced",
    "lbfgsb_unbalanced",
    "lbfgsb_unbalanced2",
]



"""Tests for module gaussian"""

# Author: Eloi Tanguy <eloi.tanguy@u-paris>
#         Remi Flamary <remi.flamary@polytehnique.edu>
#         Julie Delon <julie.delon@math.cnrs.fr>
#
# License: MIT License

import numpy as np
import pytest
from ot.utils import proj_simplex
from ot.gmm import (
    gaussian_pdf,
    gmm_pdf,
    dist_bures_squared,
    gmm_ot_loss,
    gmm_ot_plan,
    gmm_ot_apply_map,
    gmm_ot_plan_density,
)

try:
    import torch
except ImportError:
    torch = False


def get_gmms(nx=None):
    rng = np.random.RandomState(seed=42)
    ks = 3
    kt = 5
    d = 3
    m_s = rng.randn(ks, d)
    m_t = rng.randn(kt, d)
    C_s = rng.randn(ks, d, d)
    C_s = np.matmul(C_s, np.transpose(C_s, (0, 2, 1)))
    C_t = rng.randn(kt, d, d)
    C_t = np.matmul(C_t, np.transpose(C_t, (0, 2, 1)))
    w_s = proj_simplex(rng.rand(ks))
    w_t = proj_simplex(rng.rand(kt))
    if nx is not None:
        m_s = nx.from_numpy(m_s)
        m_t = nx.from_numpy(m_t)
        C_s = nx.from_numpy(C_s)
        C_t = nx.from_numpy(C_t)
        w_s = nx.from_numpy(w_s)
        w_t = nx.from_numpy(w_t)
    return m_s, m_t, C_s, C_t, w_s, w_t


def test_gaussian_pdf(nx):
    rng = np.random.RandomState(seed=42)
    n = 7
    d = 3
    x = nx.from_numpy(rng.randn(n, d))
    m, _, C, _, _, _ = get_gmms(nx)
    pdf = gaussian_pdf(x, m[0], C[0])
    assert pdf.shape == (n,)

    x = nx.from_numpy(rng.randn(n, n, d))
    pdf = gaussian_pdf(x, m[0], C[0])
    assert pdf.shape == (
        n,
        n,
    )

    with pytest.raises(AssertionError):
        gaussian_pdf(x, m[0, :-1], C[0])


def test_gmm_pdf(nx):
    rng = np.random.RandomState(seed=42)
    n = 7
    d = 3
    x = nx.from_numpy(rng.randn(n, d))
    m, _, C, _, w, _ = get_gmms(nx)
    pdf = gmm_pdf(x, m, C, w)
    assert pdf.shape == (n,)

    x = nx.from_numpy(rng.randn(n, n, d))
    pdf = gmm_pdf(x, m, C, w)
    assert pdf.shape == (
        n,
        n,
    )

    with pytest.raises(AssertionError):
        gmm_pdf(x, m[:-1], C, w)


@pytest.skip_backend("tf")  # skips because of array assignment
@pytest.skip_backend("jax")
def test_dist_bures_squared(nx):
    m_s, m_t, C_s, C_t, _, _ = get_gmms(nx)
    dist_bures_squared(m_s, m_t, C_s, C_t)
    D0 = dist_bures_squared(m_s, m_s, C_s, C_s)

    assert np.allclose(np.diag(D0), 0, atol=1e-6)

    with pytest.raises(AssertionError):
        dist_bures_squared(m_s[:, 1:], m_t, C_s, C_t)

    with pytest.raises(AssertionError):
        dist_bures_squared(m_s[1:], m_t, C_s, C_t)

    with pytest.raises(AssertionError):
        dist_bures_squared(m_s, m_t[1:], C_s, C_t)


@pytest.skip_backend("tf")  # skips because of array assignment
@pytest.skip_backend("jax")
def test_gmm_ot_loss(nx):
    m_s, m_t, C_s, C_t, w_s, w_t = get_gmms(nx)
    loss = gmm_ot_loss(m_s, m_t, C_s, C_t, w_s, w_t)

    assert loss > 0

    loss = gmm_ot_loss(m_s, m_s, C_s, C_s, w_s, w_s)

    assert np.allclose(loss, 0, atol=1e-6)

    with pytest.raises(AssertionError):
        gmm_ot_loss(m_s, m_t, C_s, C_t, w_s[1:], w_t)

    with pytest.raises(AssertionError):
        gmm_ot_loss(m_s, m_t, C_s, C_t, w_s, w_t[1:])


@pytest.skip_backend("tf")  # skips because of array assignment
@pytest.skip_backend("jax")
def test_gmm_ot_plan(nx):
    m_s, m_t, C_s, C_t, w_s, w_t = get_gmms(nx)

    plan = gmm_ot_plan(m_s, m_t, C_s, C_t, w_s, w_t)

    assert np.allclose(plan.sum(0), w_t, atol=1e-6)
    assert np.allclose(plan.sum(1), w_s, atol=1e-6)

    plan = gmm_ot_plan(m_s, m_s + 1, C_s, C_s, w_s, w_s)

    assert np.allclose(plan, np.diag(w_s), atol=1e-6)

    with pytest.raises(AssertionError):
        gmm_ot_loss(m_s, m_t, C_s, C_t, w_s[1:], w_t)

    with pytest.raises(AssertionError):
        gmm_ot_loss(m_s, m_t, C_s, C_t, w_s, w_t[1:])


def test_gmm_apply_map():
    m_s, m_t, C_s, C_t, w_s, w_t = get_gmms()
    rng = np.random.RandomState(seed=42)
    x = rng.randn(7, 3)

    for method in ["bary", "rand"]:
        gmm_ot_apply_map(x, m_s, m_t, C_s, C_t, w_s, w_t, method=method)

    plan = gmm_ot_plan(m_s, m_t, C_s, C_t, w_s, w_t)
    gmm_ot_apply_map(x, m_s, m_t, C_s, C_t, w_s, w_t, plan=plan)


@pytest.mark.skipif(not torch, reason="No torch available")
def test_gradient_gmm_ot_loss_pytorch():
    m_s, m_t, C_s, C_t, w_s, w_t = get_gmms()
    m_s = torch.tensor(m_s, requires_grad=True)
    m_t = torch.tensor(m_t, requires_grad=True)
    C_s = torch.tensor(C_s, requires_grad=True)
    C_t = torch.tensor(C_t, requires_grad=True)
    w_s = torch.tensor(w_s, requires_grad=True)
    w_t = torch.tensor(w_t, requires_grad=True)
    loss = gmm_ot_loss(m_s, m_t, C_s, C_t, w_s, w_t)
    loss.backward()
    grad_m_s = m_s.grad
    grad_C_s = C_s.grad
    grad_w_s = w_s.grad
    assert (grad_m_s**2).sum().item() > 0
    assert (grad_C_s**2).sum().item() > 0
    assert (grad_w_s**2).sum().item() > 0


def test_gmm_ot_plan_density(nx):
    m_s, m_t, C_s, C_t, w_s, w_t = get_gmms(nx)
    rng = np.random.RandomState(seed=42)
    n = 7
    x = nx.from_numpy(rng.randn(n, 3))
    y = nx.from_numpy(rng.randn(n + 1, 3))

    density = gmm_ot_plan_density(x, y, m_s, m_t, C_s, C_t, w_s, w_t)
    assert density.shape == (n, n + 1)

    plan = gmm_ot_plan(m_s, m_t, C_s, C_t, w_s, w_t)
    gmm_ot_plan_density(x, x, m_s, m_t, C_s, C_t, w_s, w_t, plan=plan)

    with pytest.raises(AssertionError):
        gmm_ot_plan_density(x[:, 1:], y, m_s, m_t, C_s, C_t, w_s, w_t)



"""Tests for ot solvers"""

# Author: Remi Flamary <remi.flamary@polytechnique.edu>
#
# License: MIT License

import itertools
import numpy as np
import pytest
import sys

import ot
from ot.bregman import geomloss
from ot.backend import torch


lst_reg = [None, 1]
lst_reg_type = ["KL", "entropy", "L2", "tuple"]
lst_unbalanced = [None, 0.9]
lst_unbalanced_type = ["KL", "L2", "TV"]

lst_reg_type_gromov = ["entropy"]
lst_gw_losses = ["L2", "KL"]
lst_unbalanced_type_gromov = ["KL", "semirelaxed", "partial"]
lst_unbalanced_gromov = [None, 0.9]
lst_alpha = [0, 0.4, 0.9, 1]

lst_method_params_solve_sample = [
    {"method": "1d"},
    {"method": "1d", "metric": "euclidean"},
    {"method": "gaussian"},
    {"method": "gaussian", "reg": 1},
    {"method": "factored", "rank": 10},
    {"method": "lowrank", "rank": 10},
]

lst_parameters_solve_sample_NotImplemented = [
    {"method": "1d", "metric": "any other one"},  # fail 1d on weird metrics
    {
        "method": "gaussian",
        "metric": "euclidean",
    },  # fail gaussian on metric not euclidean
    {
        "method": "factored",
        "metric": "euclidean",
    },  # fail factored on metric not euclidean
    {
        "method": "lowrank",
        "metric": "euclidean",
    },  # fail lowrank on metric not euclidean
    {"lazy": True},  # fail lazy for non regularized
    {"lazy": True, "unbalanced": 1},  # fail lazy for non regularized unbalanced
    {
        "lazy": True,
        "reg": 1,
        "unbalanced": 1,
    },  # fail lazy for unbalanced and regularized
]

# set readable ids for each param
lst_method_params_solve_sample = [
    pytest.param(param, id=str(param)) for param in lst_method_params_solve_sample
]
lst_parameters_solve_sample_NotImplemented = [
    pytest.param(param, id=str(param))
    for param in lst_parameters_solve_sample_NotImplemented
]


def assert_allclose_sol(sol1, sol2):
    lst_attr = [
        "value",
        "value_linear",
        "plan",
        "potential_a",
        "potential_b",
        "marginal_a",
        "marginal_b",
    ]

    nx1 = sol1._backend if sol1._backend is not None else ot.backend.NumpyBackend()
    nx2 = sol2._backend if sol2._backend is not None else ot.backend.NumpyBackend()

    for attr in lst_attr:
        if getattr(sol1, attr) is not None and getattr(sol2, attr) is not None:
            try:
                np.allclose(
                    nx1.to_numpy(getattr(sol1, attr)),
                    nx2.to_numpy(getattr(sol2, attr)),
                    equal_nan=True,
                )
            except NotImplementedError:
                pass
        elif getattr(sol1, attr) is None and getattr(sol2, attr) is None:
            return True
        else:
            return False


def test_solve(nx):
    n_samples_s = 10
    n_samples_t = 7
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples_s, n_features)
    y = rng.randn(n_samples_t, n_features)
    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    M = ot.dist(x, y)
    reg = 1e-1

    # solve unif weights
    sol0 = ot.solve(M, reg=reg)

    print(sol0)

    # solve signe weights
    sol = ot.solve(M, a, b, reg=reg)

    # check some attributes
    sol.potentials
    sol.sparse_plan
    sol.marginals
    sol.status

    # print("dual = {}".format(sol.potentials))
    # assert_allclose_sol(sol0, sol)

    # solve in backend
    ab, bb, Mb = nx.from_numpy(a, b, M)
    solb = ot.solve(M, a, b)

    assert_allclose_sol(sol, solb)

    # test not implemented unbalanced and check raise
    with pytest.raises(NotImplementedError):
        sol0 = ot.solve(M, unbalanced=1, unbalanced_type="cryptic divergence")

    # test not implemented reg_type and check raise
    with pytest.raises(NotImplementedError):
        sol0 = ot.solve(M, reg=1, reg_type="cryptic divergence")


@pytest.mark.skipif(not torch, reason="torch no installed")
def test_solve_envelope():
    n_samples_s = 10
    n_samples_t = 7
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples_s, n_features)
    y = rng.randn(n_samples_t, n_features)
    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)
    M = ot.dist(x, y)

    a = torch.tensor(a, requires_grad=True)
    b = torch.tensor(b, requires_grad=True)
    M = torch.tensor(M, requires_grad=True)

    sol0 = ot.solve(M, a, b, reg=10, grad="envelope")
    sol0.value.backward()

    gM0 = M.grad.clone()
    ga0 = a.grad.clone()
    gb0 = b.grad.clone()

    a = torch.tensor(a, requires_grad=True)
    b = torch.tensor(b, requires_grad=True)
    M = torch.tensor(M, requires_grad=True)

    sol = ot.solve(M, a, b, reg=10, grad="autodiff")
    sol.value.backward()

    gM = M.grad.clone()
    ga = a.grad.clone()
    gb = b.grad.clone()

    # Note, gradients aer invariant to change in constant so we center them
    assert torch.allclose(gM0, gM)
    assert torch.allclose(ga0 - ga0.mean(), ga - ga.mean())
    assert torch.allclose(gb0 - gb0.mean(), gb - gb.mean())


@pytest.mark.parametrize(
    "reg,reg_type,unbalanced,unbalanced_type",
    itertools.product(lst_reg, lst_reg_type, lst_unbalanced, lst_unbalanced_type),
)
def test_solve_grid(nx, reg, reg_type, unbalanced, unbalanced_type):
    n_samples_s = 10
    n_samples_t = 7
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples_s, n_features)
    y = rng.randn(n_samples_t, n_features)
    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    M = ot.dist(x, y)

    try:
        if reg_type == "tuple":

            def f(G):
                return np.sum(G**2)

            def df(G):
                return 2 * G

            reg_type = (f, df)

        # solve unif weights
        sol0 = ot.solve(
            M,
            reg=reg,
            reg_type=reg_type,
            unbalanced=unbalanced,
            unbalanced_type=unbalanced_type,
        )

        # solve signe weights
        sol = ot.solve(
            M,
            a,
            b,
            reg=reg,
            reg_type=reg_type,
            unbalanced=unbalanced,
            unbalanced_type=unbalanced_type,
        )

        assert_allclose_sol(sol0, sol)

        # solve in backend
        ab, bb, Mb = nx.from_numpy(a, b, M)

        if isinstance(reg_type, tuple):

            def f(G):
                return nx.sum(G**2)

            def df(G):
                return 2 * G

            reg_type = (f, df)

        solb = ot.solve(
            Mb,
            ab,
            bb,
            reg=reg,
            reg_type=reg_type,
            unbalanced=unbalanced,
            unbalanced_type=unbalanced_type,
        )

        assert_allclose_sol(sol, solb)

    except NotImplementedError:
        pytest.skip("Not implemented")


def test_solve_not_implemented(nx):
    n_samples_s = 10
    n_samples_t = 7
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples_s, n_features)
    y = rng.randn(n_samples_t, n_features)

    M = ot.dist(x, y)

    # test not implemented and check raise
    with pytest.raises(NotImplementedError):
        ot.solve(M, reg=1.0, reg_type="cryptic divergence")
    with pytest.raises(NotImplementedError):
        ot.solve(M, unbalanced=1.0, unbalanced_type="cryptic divergence")


def test_solve_gromov(nx):
    np.random.seed(0)

    n_samples_s = 3
    n_samples_t = 5

    Ca = np.random.rand(n_samples_s, n_samples_s)
    Ca = (Ca + Ca.T) / 2

    Cb = np.random.rand(n_samples_t, n_samples_t)
    Cb = (Cb + Cb.T) / 2

    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    M = np.random.rand(n_samples_s, n_samples_t)

    sol0 = ot.solve_gromov(Ca, Cb)  # GW
    sol = ot.solve_gromov(Ca, Cb, a=a, b=b)  # GW
    sol0_fgw = ot.solve_gromov(Ca, Cb, M)  # FGW

    # check some attributes
    sol.potentials
    sol.marginals

    assert_allclose_sol(sol0, sol)

    # solve in backend
    ax, bx, Mx, Cax, Cbx = nx.from_numpy(a, b, M, Ca, Cb)

    solx = ot.solve_gromov(Cax, Cbx, a=ax, b=bx)  # GW
    solx_fgw = ot.solve_gromov(Cax, Cbx, Mx)  # FGW

    assert_allclose_sol(sol, solx)
    assert_allclose_sol(sol0_fgw, solx_fgw)


@pytest.mark.parametrize(
    "reg,reg_type,unbalanced,unbalanced_type,alpha,loss",
    itertools.product(
        lst_reg,
        lst_reg_type_gromov,
        lst_unbalanced_gromov,
        lst_unbalanced_type_gromov,
        lst_alpha,
        lst_gw_losses,
    ),
)
def test_solve_gromov_grid(nx, reg, reg_type, unbalanced, unbalanced_type, alpha, loss):
    np.random.seed(0)

    n_samples_s = 3
    n_samples_t = 5

    Ca = np.random.rand(n_samples_s, n_samples_s)
    Ca = (Ca + Ca.T) / 2

    Cb = np.random.rand(n_samples_t, n_samples_t)
    Cb = (Cb + Cb.T) / 2

    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    M = np.random.rand(n_samples_s, n_samples_t)

    try:
        sol0 = ot.solve_gromov(
            Ca,
            Cb,
            reg=reg,
            reg_type=reg_type,
            unbalanced=unbalanced,
            unbalanced_type=unbalanced_type,
            loss=loss,
        )  # GW
        sol0_fgw = ot.solve_gromov(
            Ca,
            Cb,
            M,
            reg=reg,
            reg_type=reg_type,
            unbalanced=unbalanced,
            unbalanced_type=unbalanced_type,
            alpha=alpha,
            loss=loss,
        )  # FGW

        # solve in backend
        ax, bx, Mx, Cax, Cbx = nx.from_numpy(a, b, M, Ca, Cb)

        solx = ot.solve_gromov(
            Cax,
            Cbx,
            reg=reg,
            reg_type=reg_type,
            unbalanced=unbalanced,
            unbalanced_type=unbalanced_type,
            loss=loss,
        )  # GW
        solx_fgw = ot.solve_gromov(
            Cax,
            Cbx,
            Mx,
            reg=reg,
            reg_type=reg_type,
            unbalanced=unbalanced,
            unbalanced_type=unbalanced_type,
            alpha=alpha,
            loss=loss,
        )  # FGW

        solx.value_quad

        assert_allclose_sol(sol0, solx)
        assert_allclose_sol(sol0_fgw, solx_fgw)

    except NotImplementedError:
        pytest.skip("Not implemented")


def test_solve_gromov_not_implemented(nx):
    np.random.seed(0)

    n_samples_s = 3
    n_samples_t = 5

    Ca = np.random.rand(n_samples_s, n_samples_s)
    Ca = (Ca + Ca.T) / 2

    Cb = np.random.rand(n_samples_t, n_samples_t)
    Cb = (Cb + Cb.T) / 2

    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    M = np.random.rand(n_samples_s, n_samples_t)

    Ca, Cb, M, a, b = nx.from_numpy(Ca, Cb, M, a, b)

    # test not implemented and check raise
    with pytest.raises(NotImplementedError):
        ot.solve_gromov(Ca, Cb, loss="weird loss")
    with pytest.raises(NotImplementedError):
        ot.solve_gromov(Ca, Cb, unbalanced=1, unbalanced_type="cryptic divergence")
    with pytest.raises(NotImplementedError):
        ot.solve_gromov(Ca, Cb, reg=1, reg_type="cryptic divergence")

    # detect partial not implemented and error detect in value
    with pytest.raises(ValueError):
        ot.solve_gromov(Ca, Cb, unbalanced_type="partial", unbalanced=1.5)
    with pytest.raises(NotImplementedError):
        ot.solve_gromov(Ca, Cb, M, unbalanced_type="partial", unbalanced=0.5)
    with pytest.raises(ValueError):
        ot.solve_gromov(Ca, Cb, reg=1, unbalanced_type="partial", unbalanced=1.5)


def test_solve_sample(nx):
    # test solve_sample when is_Lazy = False
    n = 20
    X_s = np.reshape(1.0 * np.arange(n), (n, 1))
    X_t = np.reshape(1.0 * np.arange(0, n), (n, 1))

    a = ot.utils.unif(X_s.shape[0])
    b = ot.utils.unif(X_t.shape[0])

    M = ot.dist(X_s, X_t)

    # solve with ot.solve
    sol00 = ot.solve(M, a, b)

    # solve unif weights
    sol0 = ot.solve_sample(X_s, X_t)

    # solve signe weights
    sol = ot.solve_sample(X_s, X_t, a, b)

    # check some attributes
    sol.potentials
    sol.sparse_plan
    sol.marginals
    sol.status

    assert_allclose_sol(sol0, sol)
    assert_allclose_sol(sol0, sol00)

    # solve in backend
    X_sb, X_tb, ab, bb = nx.from_numpy(X_s, X_t, a, b)
    solb = ot.solve_sample(X_sb, X_tb, ab, bb)

    assert_allclose_sol(sol, solb)

    # test not implemented unbalanced and check raise
    with pytest.raises(NotImplementedError):
        sol0 = ot.solve_sample(
            X_s, X_t, unbalanced=1, unbalanced_type="cryptic divergence"
        )

    # test not implemented reg_type and check raise
    with pytest.raises(NotImplementedError):
        sol0 = ot.solve_sample(X_s, X_t, reg=1, reg_type="cryptic divergence")


def test_solve_sample_lazy(nx):
    # test solve_sample when is_Lazy = False
    n = 20
    X_s = np.reshape(1.0 * np.arange(n), (n, 1))
    X_t = np.reshape(1.0 * np.arange(0, n), (n, 1))

    a = ot.utils.unif(X_s.shape[0])
    b = ot.utils.unif(X_t.shape[0])

    X_s, X_t, a, b = nx.from_numpy(X_s, X_t, a, b)

    M = ot.dist(X_s, X_t)

    # solve with ot.solve
    sol00 = ot.solve(M, a, b, reg=1)

    sol0 = ot.solve_sample(X_s, X_t, a, b, reg=1)

    # solve signe weights
    sol = ot.solve_sample(X_s, X_t, a, b, reg=1, lazy=True)

    assert_allclose_sol(sol0, sol00)

    np.testing.assert_allclose(sol0.plan, sol.lazy_plan[:], rtol=1e-5, atol=1e-5)


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
@pytest.mark.skipif(not geomloss, reason="pytorch not installed")
@pytest.skip_backend("tf")
@pytest.skip_backend("cupy")
@pytest.skip_backend("jax")
@pytest.mark.parametrize("metric", ["sqeuclidean", "euclidean"])
def test_solve_sample_geomloss(nx, metric):
    # test solve_sample when is_Lazy = False
    n_samples_s = 13
    n_samples_t = 7
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples_s, n_features)
    y = rng.randn(n_samples_t, n_features)
    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    xb, yb, ab, bb = nx.from_numpy(x, y, a, b)

    sol0 = ot.solve_sample(xb, yb, ab, bb, reg=1)

    # solve signe weights
    sol = ot.solve_sample(xb, yb, ab, bb, reg=1, method="geomloss")
    assert_allclose_sol(sol0, sol)

    sol1 = ot.solve_sample(xb, yb, ab, bb, reg=1, lazy=False, method="geomloss")
    assert_allclose_sol(sol0, sol)

    sol1 = ot.solve_sample(
        xb, yb, ab, bb, reg=1, lazy=True, method="geomloss_tensorized"
    )
    np.testing.assert_allclose(
        nx.to_numpy(sol1.lazy_plan[:]),
        nx.to_numpy(sol.lazy_plan[:]),
        rtol=1e-5,
        atol=1e-5,
    )

    sol1 = ot.solve_sample(xb, yb, ab, bb, reg=1, lazy=True, method="geomloss_online")
    np.testing.assert_allclose(
        nx.to_numpy(sol1.lazy_plan[:]),
        nx.to_numpy(sol.lazy_plan[:]),
        rtol=1e-5,
        atol=1e-5,
    )

    sol1 = ot.solve_sample(
        xb, yb, ab, bb, reg=1, lazy=True, method="geomloss_multiscale"
    )
    np.testing.assert_allclose(
        nx.to_numpy(sol1.lazy_plan[:]),
        nx.to_numpy(sol.lazy_plan[:]),
        rtol=1e-5,
        atol=1e-5,
    )

    sol1 = ot.solve_sample(xb, yb, ab, bb, reg=1, lazy=True, method="geomloss")
    np.testing.assert_allclose(
        nx.to_numpy(sol1.lazy_plan[:]),
        nx.to_numpy(sol.lazy_plan[:]),
        rtol=1e-5,
        atol=1e-5,
    )


@pytest.mark.parametrize("method_params", lst_method_params_solve_sample)
def test_solve_sample_methods(nx, method_params):
    n_samples_s = 20
    n_samples_t = 7
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples_s, n_features)
    y = rng.randn(n_samples_t, n_features)
    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    xb, yb, ab, bb = nx.from_numpy(x, y, a, b)

    sol = ot.solve_sample(x, y, **method_params)
    solb = ot.solve_sample(xb, yb, ab, bb, **method_params)

    # check some attributes (no need )
    assert_allclose_sol(sol, solb)

    sol2 = ot.solve_sample(x, x, **method_params)
    if method_params["method"] not in ["factored", "lowrank"]:
        np.testing.assert_allclose(sol2.value, 0)


@pytest.mark.parametrize("method_params", lst_parameters_solve_sample_NotImplemented)
def test_solve_sample_NotImplemented(nx, method_params):
    n_samples_s = 20
    n_samples_t = 7
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples_s, n_features)
    y = rng.randn(n_samples_t, n_features)
    a = ot.utils.unif(n_samples_s)
    b = ot.utils.unif(n_samples_t)

    xb, yb, ab, bb = nx.from_numpy(x, y, a, b)

    with pytest.raises(NotImplementedError):
        ot.solve_sample(xb, yb, ab, bb, **method_params)
